# Copyright 2017-2019 Tom Eulenfeld, MIT license
"""Stack correlations"""
import numpy as np
import obspy
from obspy import UTCDateTime as UTC

from yam.util import _corr_id, _time2sec, IterTime


def stack(stream, length=None, move=None):
    """
    Stack traces in stream by correlation id

    :param stream: |Stream| object with correlations
    :param length: time span of one trace in the stack in seconds
        (alternatively a string consisting of a number and a unit
        -- ``'d'`` for days and ``'h'`` for hours -- can be specified,
        i.e. ``'3d'`` stacks together all traces inside a three days time
        window, default: None, which stacks together all traces)
    :param move: define a moving stack, float or string,
        default: None -- no moving stack,
        if specified move usually is smaller than length to get an overlap
        in the stacked traces
    :return: |Stream| object with stacked correlations
    """
    stream.sort()
    stream_stack = obspy.Stream()
    ids = {_corr_id(tr) for tr in stream}
    ids.discard(None)
    for id_ in ids:
        traces = [tr for tr in stream if _corr_id(tr) == id_]
        if length is None:
            data = np.mean([tr.data for tr in traces], dtype=float, axis=0)
            tr_stack = obspy.Trace(data, header=traces[0].stats)
            tr_stack.stats.key = tr_stack.stats.key + '_s'
            if 'num' in traces[0].stats:
                tr_stack.stats.num = sum(tr.stats.num for tr in traces)
            else:
                tr_stack.stats.num = len(traces)
            stream_stack.append(tr_stack)
        else:
            t1 = traces[0].stats.starttime
            lensec = _time2sec(length)
            movesec = _time2sec(move) if move else lensec
            if (lensec % (24 * 3600) == 0 or
                    isinstance(length, str) and 'd' in length):
                t1 = UTC(t1.year, t1.month, t1.day)
            elif (lensec % 3600 == 0 or
                    isinstance(length, str) and 'm' in length):
                t1 = UTC(t1.year, t1.month, t1.day, t1.hour)
            t2 = max(t1, traces[-1].stats.endtime - lensec)
            for t in IterTime(t1, t2, dt=movesec):
                sel = [tr for tr in traces
                       if -0.1 <= tr.stats.starttime - t <= lensec + 0.1]
                if len(sel) == 0:
                    continue
                data = np.mean([tr.data for tr in sel], dtype=float, axis=0)
                tr_stack = obspy.Trace(data, header=sel[0].stats)
                key_add = '_s%s' % length + (move is not None) * ('m%s' % move)
                tr_stack.stats.key = tr_stack.stats.key + key_add
                tr_stack.stats.starttime = t
                if 'num' in traces[0].stats:
                    tr_stack.stats.num = sum(tr.stats.num for tr in sel)
                else:
                    tr_stack.stats.num = len(sel)
                stream_stack.append(tr_stack)
    return stream_stack



# Copyright Thomas Lecocq, originally European Union Public Licence V. 1.1
# from msnoise.api

import numpy as np
from scipy.fftpack import fft, ifft, fftfreq, next_fast_len
import logging

log = logging.getLogger('yam.correlate')


def check_and_phase_shift(trace):
    if trace.stats.npts < 10 * trace.stats.sampling_rate:
        trace.data = np.zeros(trace.stats.npts)
        return
    dt = np.mod(trace.stats.starttime.datetime.microsecond * 1.0e-6,
                trace.stats.delta)
    if (trace.stats.delta - dt) <= np.finfo(float).eps:
        dt = 0.
    if dt != 0.:
        if dt <= (trace.stats.delta / 2.):
            dt = -dt
        else:
            dt = (trace.stats.delta - dt)
        log.debug("correcting time of trace %s with starttime %s by %.6fs",
                  trace.id, trace.stats.starttime, dt)
        trace.detrend(type="demean")
        trace.detrend(type="simple")
        trace.taper(max_percentage=None, max_length=1.0)

        n = next_fast_len(int(trace.stats.npts))
        FFTdata = fft(trace.data, n=n)
        freq = fftfreq(n, d=trace.stats.delta)
        FFTdata = FFTdata * np.exp(1j * 2. * np.pi * freq * dt)
        FFTdata = FFTdata.astype(np.complex64)
        ifft(FFTdata, n=n, overwrite_x=True)
        trace.data = np.real(FFTdata[:len(trace.data)]).astype(np.float)
        trace.stats.starttime += dt
#        clean_scipy_cache()



# -*- coding: utf-8 -*-
# Copyright 2017-2019 Tom Eulenfeld, MIT license
"""
Plotting functions

Common arguments in plotting functions are:

:stream: |Stream| object with correlations
:fname: file name for the plot output
:ext: file name extension (e.g. ``'png'``, ``'pdf'``)
:figsize: figure size (tuple of inches)
:dpi: resolution of image file (not available for station plot)
:xlim: limits of x axis (tuple of lag times or tuple of UTC strings)
:ylim: limits of y axis (tuple of UTC strings or tuple of percentages)
:line_style: style of a wiggle plot, see |Axes.plot| in matplotlib's documentation
:line_width: line width of wiggle plot

|
"""

from copy import copy
from collections import OrderedDict
import os.path
from warnings import warn

import matplotlib.pyplot as plt
import numpy as np
from obspy import UTCDateTime as UTC

from yam.util import _corr_id, _trim, _trim_time_period
import yam.stack


def _get_times_no_data(x):
    """Return times without data"""
    dx = np.median(np.diff(x))
    try:
        index = np.argwhere(np.diff(x) > dx)[:, 0]
    except IndexError:
        return {}
    tl = OrderedDict()
    for i in index:
        tl[i] = int(round((x[i + 1] - x[i]) / dx)) - 1
    return tl


def _add_value(x, tl, value=None, masked=False, single_value=True):
    """Add values for missing data

    :param x: The array where to add the data
    :param tl: output of `_get_times_no_data`, dictionary with times
    :param value: fill value, if None assume a time array
    :param masked: mask filled values
    :param single_value: True -- add a single value for plotting,
        False -- fill gaps fully
    """
    dx = np.median(np.diff(x))
    keys = sorted(tl.keys())
    xs = np.split(x, [i + 1 for i in keys], axis=-1)
    dtype = xs[0].dtype
    xf = [xs[0]]
    for i, x in enumerate(xs[1:]):
        if single_value:
            shape = (1,)
        else:
            shape = (tl[keys[i]],)
        if len(x.shape) > 1:
            shape = (x.shape[0],) + shape
        if value is None:
            # shape not sorted out, but doesnt matter
            # can only be used with single_value = True!!
            # if this should be supported it needs a proper test!
            if not single_value:
                raise NotImplementedError
            xfn = xf[-1][-1] + dx
        else:
            xfn = np.ones(shape, dtype=dtype) * value
        if masked:
            xfn = np.ma.masked_all_like(xfn)
        xf.append(xfn)
        xf.append(x)
    x2 = np.ma.hstack(xf)
    return x2


def _align_values_for_pcolormesh(x):
    x = list(x)
    dx = np.median(np.diff(x))
    x.append(x[-1] + dx)
    x = [xx - 0.5 * dx for xx in x]
    return x


def plot_stations(inventory, fname, ext='png', projection='local', **kwargs):
    """
    Plot station map

    :param inventory: |Inventory| object with coordinates
    :param projection,\*\*kwargs: passed to |Inventory.plot| method
    """
    inventory.plot(projection, outfile=fname + '.' + ext, **kwargs)


def plot_data(data, fname, ext='png', show=False,
              type='dayplot', **kwargs):
    """
    Plot data (typically one day)

    :param data: |Stream| object holding the data
    :param type,\*\*kwargs: passed to |Stream.plot| method
    """
    label = os.path.basename(fname)
    data.plot(type=type, outfile=fname + '.' + ext, title=label, **kwargs)
    if show:
        data.plot(type=type, title=label, show=True, **kwargs)


def plot_corr_vs_dist(
        stream, fname=None, figsize=(10, 5), ext='png', dpi=None,
        components='ZZ', line_style='k', scale=1, dist_unit='km',
        xlim=None, ylim=None, time_period=None):
    """
    Plot stacked correlations versus inter-station distance

    .. image:: _static/corr_vs_dist.png
       :width: 30%

    This plot can be created from the command line with ``--plottype vs_dist``.

    :param components: component combination to plot
    :param scale: scale wiggles (default 1)
    :param dist_unit: one of ``('km', 'm', 'deg')``
    :time_period: use correlations only from this time span (tuple of dates)
    """
    # scale relative to axis
    traces = [tr for tr in stream if
              _corr_id(tr).split('-')[0][-1] + _corr_id(tr)[-1] == components]
    stream.traces = traces
    _trim_time_period(stream, time_period)
    stack = yam.stack.stack(stream)
    if len(stack) == 0:
        msg = 'Not plotting anything. No traces in stack with components %s'
        warn(msg % components)
        return
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    dist_scale = (1 if dist_unit == 'm' else 1000 * 111.2 if dist_unit == 'deg'
                  else 1000)
    max_dist = max(tr.stats.dist / dist_scale for tr in stack)
    for tr in stack:
        lag_times = _trim(tr, xlim)
        scaled_data = tr.stats.dist / dist_scale + tr.data * max_dist * scale
        ax.plot(lag_times, scaled_data, line_style)
    if fname is not None:
        fname = '%s_%s' % (fname, components)
        label = os.path.basename(fname)
    else:
        label = components
    ax.set_ylim(ylim)
    ax.annotate(label, (0, 1), (10, 10), 'axes fraction', 'offset points',
                annotation_clip=False, va='bottom')
    ax.set_ylabel('distance (%s)' % dist_unit)
    ax.set_xlabel('time (s)')
    if fname is not None:
        fig.savefig(fname + '.' + ext, dpi=dpi)


def plot_corr_vs_time_wiggle(
        stream, fname=None, figsize=(10, 5), ext='png', dpi=None,
        xlim=None, ylim=None,
        line_style='k', line_width=0.5, scale=20):
    """
    Plot correlation wiggles versus time

    .. image:: _static/corr_vs_time_wiggle.png
       :width: 30%
    .. image:: _static/corr_vs_time_wiggle2.png
       :width: 40%

    This plot can be created from the command line with ``--plottype wiggle``.

    :param scale: scale of wiggles (default 20)
    """
    # scale relative to neighboring wiggles
    ids = {_corr_id(tr) for tr in stream}
    if len(ids) != 1:
        warn('Different ids in stream: %s' % ids)
    stream.sort(['starttime'])
    _trim_time_period(stream, ylim)
    times = [tr.stats.starttime.matplotlib_date for tr in stream]
    dt = np.median(np.diff(times))
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    for tr in stream:
        lag_times = _trim(tr, xlim)
        scaled_data = tr.stats.starttime.matplotlib_date + tr.data * dt * scale
        ax.plot(lag_times, scaled_data, line_style, lw=line_width)
    label = '' if fname is None else os.path.basename(fname)
    ax.annotate(label, (0, 1), (10, 10), 'axes fraction', 'offset points',
                annotation_clip=False, va='bottom')
    ax.set_ylabel('date')
    ax.set_xlabel('time (s)')
    ax.yaxis_date()
    if fname is not None:
        fig.savefig(fname + '.' + ext, dpi=dpi)


def plot_corr_vs_time(
        stream, fname=None, figsize=(10, 5), ext='png', dpi=None,
        xlim=None, ylim=None, vmax=None, cmap='RdBu_r',
        show_stack=True, line_style='k', line_width=1):
    """
    Plot correlations versus time

    .. image:: _static/corr_vs_time.png
       :width: 30%
    .. image:: _static/corr_vs_time_zoom.png
       :width: 30%

    Default correlation plot.

    :param vmax: maximum value in colormap
    :param cmap: used colormap
    :param show_stack: show a wiggle plot of the stack at top
    """
    ids = {_corr_id(tr) for tr in stream}
    srs = {tr.stats.sampling_rate for tr in stream}
    lens = {len(tr) for tr in stream}
    if len(ids) != 1:
        warn('Different ids in stream: %s' % ids)
    # These checks should be done when saving the correlations to hdf5.
    # On the other hand side, these problems will not occur often.
    if len(srs) != 1:
        sr = np.median([tr.stats.sampling_rate for tr in stream])
        msg = 'Different sampling rates in stream: %s -> set %s Hz (%s)'
        warn(msg % (srs, sr, stream[0].id))
        for tr in stream:
            tr.stats.sampling_rate = sr
    if len(lens) != 1:
        msg = ('Different lengths of traces in stream: %s (%s)'
               '-> Plese use xlim parameter to trim traces')
        warn(msg % (lens, stream[0].id))
    stream.sort(['starttime'])
    _trim_time_period(stream, ylim)
    for tr in stream:
        lag_times = _trim(tr, xlim)
    times = [tr.stats.starttime.matplotlib_date for tr in stream]
    fig = plt.figure(figsize=figsize)
    ax = fig.add_axes([0.15, 0.1, 0.75, 0.75])
    cax = fig.add_axes([0.91, 0.375, 0.008, 0.25])
    data = np.array([tr.data for tr in stream])
    if vmax is None:
        vmax = min(0.8 * np.max(data), 0.1)
    no_data = _get_times_no_data(times)
    times = _add_value(times, no_data)
    times2 = _align_values_for_pcolormesh(times)
    lag_times2 = _align_values_for_pcolormesh(lag_times)
    data2 = _add_value(np.transpose(data), no_data, value=0, masked=True)
    data2 = np.transpose(data2)
    mesh = ax.pcolormesh(lag_times2, times2, data2, cmap=cmap,
                         vmin=-vmax, vmax=vmax)
    fig.colorbar(mesh, cax)
    ax.set_ylabel('date')
    ax.set_xlabel('time (s)')
    ax.yaxis_date()
    ax_ano = ax
    if show_stack:
        ax2 = fig.add_axes([0.15, 0.85, 0.75, 0.05], sharex=ax)
        stack = yam.stack.stack(stream)
        ax2.plot(lag_times, stack[0].data, line_style, lw=line_width)
        ax2.set_ylim(-vmax, vmax)
        ax2.set_ylabel('stack')
        plt.setp(ax2.get_xticklabels(), visible=False)
        ax_ano = ax2
    ax.set_xlim(lag_times[0], lag_times[-1])
    label = '' if fname is None else os.path.basename(fname)
    ax_ano.annotate(label, (0, 1), (10, 10), 'axes fraction', 'offset points',
                    annotation_clip=False, va='bottom')
    if fname is not None:
        fig.savefig(fname + '.' + ext, dpi=dpi)


def plot_sim_mat(res, fname=None, figsize=(10, 5), ext='png', dpi=None,
                 xlim=None, ylim=None, vmax=None, cmap='hot_r',
                 show_line=False, line_style='b', line_width=2,
                 time_window=None):
    """
    Plot similarity matrices

    .. image:: _static/sim_mat.png
       :width: 30%

    Default plot for stretching results.

    :param res: dictionary with stretching results
    :param vmax: maximum value in colormap
    :param cmap: used colormap
    :param show_line: show line connecting best correlations for each time
    :param time_window: do not create figures for each time window in the
        results dictionary, but only for one time window with given index
    """
    tw = res['tw']
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    data = np.transpose(res['sim_mat'])
    x = [UTC(t) for t in res['times']]
    no_data = _get_times_no_data(x)
    x = _add_value(x, no_data)
    x = [t.datetime for t in x]
    x2 = _align_values_for_pcolormesh(x)
    y2 = _align_values_for_pcolormesh(copy(res['velchange_values']))
    if vmax is None:
        vmax = np.max(data)
    data = _add_value(data, no_data, value=0, masked=True)
    mesh = ax.pcolormesh(x2, y2, data, cmap=cmap, vmin=0, vmax=vmax)
    if show_line:
        s = res['velchange_vs_time']
        s = _add_value(s, no_data, value=0, masked=True)
        ax.plot(x, s, line_style, lw=line_width)
    ax.set_xlabel('date')
    ax.set_ylabel('velocity change (%)')
    fig.autofmt_xdate()
    fig.colorbar(mesh, shrink=0.5)
    label_tw = 'tw_{:05.1f}s-{:05.1f}s'.format(*tw)
    if fname is None:
        label = label_tw
    else:
        label = os.path.basename(fname) + '_' + label_tw
    ax.annotate(label, (0, 1), (10, 10), 'axes fraction', 'offset points',
                annotation_clip=False, va='bottom')
    if ylim:
        if isinstance(ylim, (float, int)):
            ylim = (-ylim, ylim)
        ax.set_ylim(ylim)
    t0, t1 = (None, None) if xlim is None else xlim
    t0 = x2[0] if t0 is None else UTC(t0).matplotlib_date
    t1 = x2[-1] if t1 is None else UTC(t1).matplotlib_date
    ax.set_xlim(t0, t1)
    if fname is not None:
        fig.savefig(fname + '.' + ext, dpi=dpi)
    return fig



### file miic.core.stretch_mod adapted by TE ###
"""
@author:
Eraldo Pomponi
@copyright:
The MIIC Development Team (eraldo.pomponi@uni-leipzig.de)
@license:
GNU Lesser General Public License, Version 3
(http://www.gnu.org/copyleft/lesser.html)
Created on Aug 19, 2011
"""
import logging

# Main imports
#import scipy.ndimage
import numpy as np
from scipy.interpolate import UnivariateSpline
from obspy.signal.invsim import cosine_taper

log = logging.getLogger('yam.stretch')

## ETS imports
#try:
#    BC_UI = True
#    from traits.api import HasTraits, Int, Float, Bool, List, Enum
#    from traitsui.api import View, Item, Tabbed
#except ImportError:
#    BC_UI = False
#    pass
#
#
##from miic.core.miic_utils import nextpow2, from_str_to_datetime, dv_check, \
##    flatten_recarray, convert_time
#
#
#def _stretch_mat_creation(ref_tr, str_range=0.1, nstr=100):
#    """ Matrix of stretched instance of a reference trace.
#    The reference trace is stretched using a cubic spline interpolation
#    algorithm form ``-str_range`` to ``str_range`` (in %) for totally
#    ``nstr`` steps.
#    The output of this function is a matrix containing the stretched version
#    of the reference trace (one each row) (``new_m``) and the corresponding
#    stretching amount (``deltas``).
#    :type ref_tr: :class:`~numpy.ndarray`
#    :param ref_tr: 1d ndarray. The reference trace that will be stretched
#    :type str_range: float
#    :param str_range: Amount, in percent, of the desired stretching (one side)
#    :type nstr: int
#    :param nstr: Number of stretching steps (one side)
#    :rtype: :class:`~numpy.ndarray`, float
#    :return:
#        **new_m**:
#            2d ndarray of stretched version of the reference trace.
#            Its size is ``(nstr,len(ref_tr)/2)`` if ``signle_side==True``
#            otherwise it is ``(nstr,len(ref_tr))``
#        **deltas**: List of float, stretch amount for each row
#            of ``new_m``
#    """
#
#    n = len(ref_tr)
#    samples_idx = np.arange(n) - n // 2
#    deltas = 1 + np.linspace(-str_range, str_range, nstr)
#    str_timemat = np.zeros((nstr, n))
#    for ii in np.arange(nstr):
#        str_timemat[ii, :] = samples_idx / deltas[nstr - 1 - ii]
#    new_m = np.zeros_like(str_timemat)
#    coord = np.zeros((2, n))
#    for (i, row) in enumerate(str_timemat):
#        coord[0, :] = row + n // 2
#        new_m[i, :] = scipy.ndimage.map_coordinates(\
#                                    ref_tr.reshape((len(ref_tr), 1)), coord)
#    return new_m, deltas
#
#
#if BC_UI:
#    class _stretch_mat_creation_view(HasTraits):
#        str_range = Float(0.1)
#        nstr = Int(100)
#        single_side = Bool(False)
#
#        trait_view = View(Item('str_range'),
#                          Item('nstr'),
#                          Item('single_side'))


def velocity_change_estimete(mat, tw, strrefmat, strvec, sides='both',
                             return_sim_mat=False, remove_nans=True):
    """ Velocity change estimate through stretching and comparison.
    Velocity changes are estimated comparing each correlation function stored
    in the ``mat`` matrix (one for each row) with ``strrefmat.shape[0]``
    stretched versions  of a reference trace stored in ``strrefmat``.
    The stretch amount for each row of ``strrefmat`` must be passed in the
    ``strvec`` vector.
    The best match (stretch amount and corresponding correlation value) is
    calculated on different time windows (each row of ``tw`` is a different
    one) instead of on the whole trace.
    :type mat: :class:`~numpy.ndarray`
    :param mat: 2d ndarray containing the correlation functions.
        One for each row.
    :type tw: :class:`~numpy.ndarray` of int
    :param tw: 2d ndarray of time windows to be use in the velocity change
         estimate.
    :type strrefmat: :class:`~numpy.ndarray`
    :param strrefmat: 2D array containing stretched version of the reference
         matrix
    :type strvec: :class:`~numpy.ndarray` or list
    :param strvec: Stretch amount for each row of ``strrefmat``
    :type sides: string
    :param sides: Side of the reference matrix to be used for the velocity
        change estimate ('both' | 'left' | 'right' | 'single')
    :type remove_nans: bool
    :param remove_nans: If `True` applay :func:`~numpy.nan_to_num` to the
        given correlation matrix before any other operation.
    :rtype: Dictionary
    :return: **dv**: Dictionary with the following keys
        *corr*: 2d ndarray containing the correlation value for the best
            match for each row of ``mat`` and for each time window.
            Its dimension is: :func:(len(tw),mat.shape[1])
        *value*: 2d ndarray containing the stretch amount corresponding to
            the best match for each row of ``mat`` and for each time window.
            Its dimension is: :func:(len(tw),mat.shape[1])
        *sim_mat*: 3d ndarray containing the similarity matricies that
            indicate the correlation coefficient with the reference for the
            different time windows, different times and different amount of
            stretching.
            Its dimension is: :py:func:`(len(tw),mat.shape[1],len(strvec))`
        *second_axis*: It contains the stretch vector used for the velocity
            change estimate.
        *vale_type*: It is equal to 'stretch' and specify the content of
            the returned 'value'.
        *method*: It is equal to 'single_ref' and specify in which "way" the
            values have been obtained.
    """

    # Mat must be a 2d vector in every case so
    mat = np.atleast_2d(mat)

    assert(strrefmat.shape[1] == mat.shape[1])

    if remove_nans:
        mat = np.nan_to_num(mat)

    center_p = np.floor((mat.shape[1] - 1.) / 2.)

    nstr = strrefmat.shape[0]

    corr = np.zeros((len(tw), mat.shape[0]))
    dt = np.zeros((len(tw), mat.shape[0]))
    sim_mat = np.zeros([mat.shape[0], len(strvec), len(tw)])

    for (ii, ctw) in enumerate(tw):

        if sides == 'both':
            ctw = np.hstack((center_p - ctw[::-1],
                             center_p + ctw)).astype(np.int32)
        elif sides == 'left':
            ctw = (center_p - ctw[::-1]).astype(np.int32)
        elif sides == 'right':
            ctw = (center_p + ctw).astype(np.int32)
        elif sides == 'single':
            ctw = ctw.astype(np.int32)
        else:
            msg = 'sides = %s not a valid option. Using sides = single'
            log.warning(msg, sides)

        mask = np.zeros((mat.shape[1],))
        mask[ctw] = 1

        ref_mask_mat = np.tile(mask, (nstr, 1))
        mat_mask_mat = np.tile(mask, (mat.shape[0], 1))

        first = mat * mat_mask_mat
        second = strrefmat * ref_mask_mat

        dprod = np.dot(first, second.T)

        # Normalization
        f_sq = np.sum(first ** 2, axis=1)
        s_sq = np.sum(second ** 2, axis=1)

        f_sq = f_sq.reshape(1, len(f_sq))
        s_sq = s_sq.reshape(1, len(s_sq))

        den = np.sqrt(np.dot(f_sq.T, s_sq))

        tmp = dprod / den
        tmp[np.isnan(tmp)] = 0  # warning is already issued
        sim_mat[:, :, ii] = tmp

#        tmp_corr_vect = tmp.max(axis=1)
#        corr[ii, :] = tmp_corr_vect
#        dt[ii, :] = strvec[tmp.argmax(axis=1)]

        # Set dt to NaN where the correlation is NaN instead of having it equal
        # to one of the two stretch_range limits
#        dt[ii, np.isnan(tmp_corr_vect)] = np.nan

    dv = {#'corr': corr,
          #'value': dt,
          'second_axis': strvec,
          'value_type': np.array(['stretch']),
          'method': np.array(['single_ref'])}

    if return_sim_mat:
        dv.update({'sim_mat': sim_mat})

    return dv


#if BC_UI:
#    class _velocity_change_estimete_view(HasTraits):
#        sides = Enum('both', 'left', 'right', 'single')
#        trait_view = View(Item('sides'))
#
#
#def time_shift_estimate(corr_data, ref_trc=None, tw=None, shift_range=10,
#                        shift_steps=100, single_sided=False,
#                        return_sim_mat=True, remove_nans=True):
#    """ Time shift estimate through shifting and comparison.
#    This function is intended to estimate shift of traces as they can occur
#    in noise cross-correlation in case of drifting clocks.
#    Time shifts are estimated comparing each correlation function stored
#    in the ``corr_data`` matrix (one for each row) with ``shift_steps``
#    shifted versions  of reference trace stored in ``ref_trc``.
#    The maximum amount of shifting may be passed in ``shift_range``.
#    The best match (shifting amount and corresponding correlation value) is
#    calculated on different time windows. If ``tw = None`` the shifting is
#    estimated on the whole trace.
#    :type corr_data: :class:`~numpy.ndarray`
#    :param corr_data: 2d ndarray containing the correlation functions.
#        One for each row.
#    :type ref_trc: :class:`~numpy.ndarray`
#    :param ref_trc: 1D array containing the reference trace to be shifted and
#        compared to the individual traces in ``mat``
#    :type tw: list of :class:`~numpy.ndarray` of int
#    :param tw: list of 1D ndarrays holding the indices of sampels in the time
#        windows to be use in the time shift estimate. The sampels are counted
#        from the zero lag time with the index of the first sample being 0. If
#        ``tw = None`` the full time range is used.
#    :type shift_range: scalar
#    :param shift_range: Maximum amount of time shift in samples (in one
#        direction).
#        Shifting is tested in both directions from ``-shift_range`` to
#        ``shift_range``
#    :type shift_steps: scalar`
#    :param shift_steps: Number of shifted version to be tested. The increment
#            will be ``(2 * shift_range) / shift_steps``
#    :type sinlge_sided: boolean
#    :param single_sided: If ``True`` the zero lag time of the traces is in the
#        first sample. If ``False`` zero lag is assumed to be in the center of
#        the traces and the shifting is evaluated on the causal and acausal
#        parts of the traces separately and averaged. This is done to avoid bias
#        from velocity changes (stretching) in the case of strongly asymmetric
#        traces.
#    :type remove_nans: bool
#    :param remove_nans: If `True` applay :func:`~numpy.nan_to_num` to the
#        given correlation matrix before any other operation.
#    :rtype: Dictionary
#    :return: **dt**: Dictionary with the following keys
#        *corr*: :class:`~numpy.ndarray` 2d ndarray containing the correlation
#            value for the best match for each row of ``mat`` and for each
#            time window.
#            Its dimension is: :func:(len(tw),mat.shape[1])
#        *value*: :class:`~numpy.ndarray` 2d ndarray containing the amount of
#            shifting corresponding to the best match for each row of ``mat``
#            and for each time window. Shift is measured in units of the
#            sampling interval.
#            Its dimension is: :py:func:`(len(tw),mat.shape[1])`
#        *sim_mat*: 2d ndarray containing the similarity matrix that
#            indicate the time shift respect to the reference for the
#            selected time windows, for different times and different amount of
#            shift.
#            Its dimension is: :py:func:`(mat.shape[1],shift_steps)`
#        *second_axis*: It contains the shift vector used for the velocity
#            change estimate.
#        *vale_type*: It is equal to 'shift' and specify the content of
#            the returned 'value'.
#        *method*: It is equal to 'time_shift' and specify in which "way" the
#            values have been obtained.
#    """
#
#    # Mat must be a 2d vector in every case so
#    mat = np.atleast_2d(corr_data)
#
#    # FIX: Trace back this problem to the original source and remove
#    # this statement
#    if remove_nans:
#        mat = np.nan_to_num(mat)
#
#    # generate the reference trace if not given (use the whole time span)
#    if ref_trc is None:
#        ref_trc = np.nansum(mat, axis=0) / mat.shape[0]
#
#    # generate time window if not given (use the full length of
#    # the correlation trace)
#    if tw is None:
#        if single_sided:
#            tw = time_windows_creation([0], [mat.shape[1]])
#        # if center is in the middle use only half the length for the
#        # time window
#        else:
#            tw = time_windows_creation([0],
#                                       [int(np.floor(mat.shape[1] / 2.))])
#
#    # taper and extend the reference trace to avoid interpolation
#    # artefacts at the ends of the trace
#    taper = cosine_taper(len(ref_trc), 0.05)
#    ref_trc *= taper
#
#    # different values of shifting to be tested
#    shifts = np.linspace(-shift_range, shift_range, shift_steps)
#
#    # time axis
#    time_idx = np.arange(len(ref_trc))
#
#    # create the array to hold the shifted traces
#    ref_shift = np.zeros((len(shifts), len(ref_trc)))
#
#    # create a spline object for the reference trace
#    ref_tr_spline = UnivariateSpline(time_idx, ref_trc, s=0)
#
#    # evaluate the spline object at different points and put in the prepared
#    # array
#    for (k, this_shift) in enumerate(shifts):
#        ref_shift[k, :] = ref_tr_spline(time_idx - this_shift)
#
#    # search best fit of the crosscorrs to one of the shifted ref_traces
#    if single_sided:
#        vdict = velocity_change_estimete(mat, tw, ref_shift,
#                                         shifts, sides='single',
#                                         return_sim_mat=True,
#                                         remove_nans=remove_nans)
#        corr = vdict['corr']
#        shift = vdict['value']
#        sim_mat = vdict['sim_mat']
#
#    else:
#        """
#        # estimate shifts for causal and acausal part individually and avarage
#        # to avoid apparent shift from velocity change and asymmetric
#        # amplitudes
#        lvdict = velocity_change_estimete(mat, tw, ref_shift,
#                                          shifts,
#                                          sides='left',
#                                          return_sim_mat=True,
#                                          remove_nans=remove_nans)
#        lcorr = lvdict['corr']
#        lshift = lvdict['value']
#        lsim_mat = lvdict['sim_mat']
#        rvdict = velocity_change_estimete(mat, tw, ref_shift,
#                                          shifts,
#                                          sides='right',
#                                          return_sim_mat=True,
#                                          remove_nans=remove_nans)
#        rcorr = rvdict['corr']
#        rshift = rvdict['value']
#        rsim_mat = rvdict['sim_mat']
#        shift = np.zeros_like(lshift)
#        corr = np.zeros_like(lshift)
#        sim_mat = np.zeros_like(lsim_mat)
#
#        corr = (lcorr + rcorr) / 2.
#        shift = (lshift + rshift) / 2.
#        sim_mat = (lsim_mat + rsim_mat) / 2.
#        """
#        dtdict = velocity_change_estimete(mat, tw, ref_shift,
#                                          shifts,
#                                          sides='both',
#                                          return_sim_mat=True,
#                                          remove_nans=remove_nans)
#        corr = dtdict['corr']
#        shift = dtdict['value']
#        sim_mat = dtdict['sim_mat']
#
#
#    # create the result dictionary
#    dt = {'corr': np.squeeze(corr),
#          'value': np.squeeze(shift),
#          'second_axis': shifts,
#          'value_type': np.array(['shift']),
#          'method': np.array(['time_shift'])}
#
#    if return_sim_mat:
#        dt.update({'sim_mat': np.squeeze(sim_mat)})
#
#    return dt
#
#
#if BC_UI:
#    class _time_shift_estimate_view(HasTraits):
#
#        shift_range = Float(10.0)
#        shift_steps = Int(100)
#        single_sided = Bool(False)
#
#        trait_view = View(Item('shift_range'),
#                          Item('shift_steps'),
#                          Item('single_sided'))
#
#
#def time_shift_apply(corr_data, shift):
#    """ Apply time shift to traces.
#    Apply time shifts to traces e.g. to align them to a common time base.
#    Such shifts can occur in correlation traces in case of a drifting clock.
#    This function ``applies`` the shifts. To correct for shift estimated with
#    :class:`~miic.core.stretch_mod.time_shift_estimate` you need to apply
#    negative shifts.
#    Shifting is done in frequency domain with 5% tapering.
#    :type corr_data: :py:class:`~numpy.ndarray`
#    :param corr_data: 2d ndarray containing the correlation functions that are
#        to be shifted.
#        One for each row.
#    :type shift: :py:class:`~numpy.ndarray`
#    :param shift: ndarray with shift.shape[0] = corr_data.shape[0] containing
#        the shifts in units of the sampling interval by which the trace are to
#        be shifted
#    :rtype: :py:class:`~numpy.ndarray`
#    :return: **shifted_mat**: shifted version of the input matrix
#    """
#
#    # Mat must be a 2d vector in every case so
#    mat = np.atleast_2d(corr_data)
#
#    # check input
#    # shift is just a 1d array
#    if len(shift.shape) == 1:
#        t_shift = np.zeros([shift.shape[0], 1])
#        t_shift[:, 0] = shift
#        shift = t_shift
#    # shift has the wrong length
#    elif shift.shape[0] != mat.shape[0]:
#        msg = 'InputError: shift.shape[0] must be equal corr_data.shape[0]'
#        log.warning(msg)
#        return 0
#    # shift has multiple columns (multiple measurements for the same time)
#    if shift.shape[1] > 1:
#        shift = np.delete(shift, np.arange(1, shift.shape[1]), axis=1)
#
#    # taper the reference matrix to avoid interpolation
#    taper = cosine_taper(mat.shape[1], 0.05)
#    mat *= np.tile(taper, [mat.shape[0], 1])
#
#    # find a suitable length for the FFT
#    N = nextpow2(2 * mat.shape[1])
#    w = np.zeros([1, N / 2 + 1])
#
#    # original and shifted phase
#    w[0, :] = np.linspace(0, np.pi, N / 2 + 1)
#    pha = np.exp(-1j * (shift) * w)
#
#    # Fourier Transform
#    F = np.fft.rfft(mat, N, 1)
#
#    # apply the phase shift
#    sF = F * pha
#
#    # transform to time domain
#    smat = np.fft.irfft(sF)
#
#    # cut to original size
#    shifted_mat = smat[:, 0:mat.shape[1]]
#    return shifted_mat
#
#
#if BC_UI:
#    class _time_shift_apply_view(HasTraits):
#        shift = List()
#        corr_data = List()
#        trait_view = View(Item('shift'), Item('corr_data'))


def time_stretch_estimate(corr_data, ref_trc=None, tw=None, stretch_range=0.1,
                          stretch_steps=100, sides='both', remove_nans=True):
    """ Time stretch estimate through stretch and comparison.
    This function estimates stretching of the time axis of traces as it can
    occur if the propagation velocity changes.
    Time stretching is estimated comparing each correlation function stored
    in the ``corr_data`` matrix (one for each row) with ``stretch_steps``
    stretched versions  of reference trace stored in ``ref_trc``.
    The maximum amount of stretching may be passed in ``stretch_range``. The
    time axis is multiplied by exp(stretch).
    The best match (stretching amount and corresponding correlation value) is
    calculated on different time windows. If ``tw = None`` the stretching is
    estimated on the whole trace.
    :type corr_data: :class:`~numpy.ndarray`
    :param corr_data: 2d ndarray containing the correlation functions.
        One for each row.
    :type ref_trc: :class:`~numpy.ndarray`
    :param ref_trc: 1D array containing the reference trace to be shifted and
        compared to the individual traces in ``mat``
    :type tw: list of :class:`~numpy.ndarray` of int
    :param tw: list of 1D ndarrays holding the indices of sampels in the time
        windows to be use in the time shift estimate. The sampels are counted
        from the zero lag time with the index of the first sample being 0. If
        ``tw = None`` the full time range is used.
    :type stretch_range: scalar
    :param stretch_range: Maximum amount of relative stretching.
        Stretching and compression is tested from ``-stretch_range`` to
        ``stretch_range``.
    :type stretch_steps: scalar`
    :param stretch_steps: Number of shifted version to be tested. The
        increment will be ``(2 * stretch_range) / stretch_steps``
    :type sides: str
    :param sides: Side of the reference matrix to be used for the stretching
        estimate ('both' | 'left' | 'right' | 'single') ``single`` is used for
        one-sided signals from active sources with zero lag time is on the
        first sample. Other options assume that the zero lag time is in the
        center of the traces.
    :type remove_nans: bool
    :param remove_nans: If `True` applay :func:`~numpy.nan_to_num` to the
        given correlation matrix before any other operation.
    :rtype: Dictionary
    :return: **dv**: Dictionary with the following keys
        *corr*: 2d ndarray containing the correlation value for the best
            match for each row of ``mat`` and for each time window.
            Its dimension is: :func:(len(tw),mat.shape[1])
        *value*: 2d ndarray containing the stretch amount corresponding to
            the best match for each row of ``mat`` and for each time window.
            Stretch is a relative value corresponding to the negative relative
            velocity change -dv/v.
            Its dimension is: :func:(len(tw),mat.shape[1])
        *sim_mat*: 3d ndarray containing the similarity matricies that
            indicate the correlation coefficient with the reference for the
            different time windows, different times and different amount of
            stretching.
            Its dimension is: :py:func:`(len(tw),mat.shape[1],len(strvec))`
        *second_axis*: It contains the stretch vector used for the velocity
            change estimate.
        *vale_type*: It is equal to 'stretch' and specify the content of
            the returned 'value'.
        *method*: It is equal to 'single_ref' and specify in which "way" the
            values have been obtained.
    """

    # Mat must be a 2d vector in every case so
    mat = np.atleast_2d(corr_data)

    # FIX: Trace back this problem to the original source and remove
    # this statement
    if remove_nans:
        mat = np.nan_to_num(mat)

    # generate the reference trace if not given (use the whole time span)
    if ref_trc is None:
        ref_trc = np.nansum(mat, axis=0) / mat.shape[0]

    # generate time window if not given (use the full length of the correlation
    # trace)
    if tw is None:
        tw = time_windows_creation([0], [int(np.floor(mat.shape[1] / 2.))])

    # taper and extend the reference trace to avoid interpolation
    # artefacts at the ends of the trace
    taper = cosine_taper(len(ref_trc), 0.05)
    ref_trc *= taper

    # different values of shifting to be tested
    stretchs = np.linspace(-stretch_range, stretch_range, stretch_steps)
    time_facs = np.exp(-stretchs)

    # time axis
    if sides is not 'single':
        time_idx = np.arange(len(ref_trc)) - (len(ref_trc) - 1.) / 2.
    else:
        time_idx = np.arange(len(ref_trc))

    # create the array to hold the shifted traces
    ref_stretch = np.zeros((len(stretchs), len(ref_trc)))

    # create a spline object for the reference trace
    ref_tr_spline = UnivariateSpline(time_idx, ref_trc, s=0)

    # evaluate the spline object at different points and put in the prepared
    # array
    for (k, this_fac) in enumerate(time_facs):
        ref_stretch[k, :] = ref_tr_spline(time_idx * this_fac)

    # search best fit of the crosscorrs to one of the stretched ref_traces
    dv = velocity_change_estimete(mat, tw, ref_stretch,
                                  stretchs, sides=sides,
                                  return_sim_mat=True,
                                  remove_nans=remove_nans)

    # TODO: It is not really clear why it it necessary to transpose here so
    # this is the fist point where to look in case of errors.
    #dv['corr'] = dv['corr'].T
    #dv['value'] = dv['value'].T
    # dv.update({'stretch_vec': stretchs})

    return dv


#if BC_UI:
#    class _time_stretch_estimate_view(HasTraits):
#
#        stretch_range = Float(0.01)
#        stretch_steps = Int(100)
#        sides = Enum('both', 'left', 'right', 'sinlge')
#
#        trait_view = View(Item('stretch_range'),
#                          Item('stretch_steps'),
#                          Item('sides'))
#
#
#def time_stretch_apply(corr_data, stretch, single_sided=False):
#    """ Apply time axis stretch to traces.
#    Stretch the time axis of traces e.g. to compensate a velocity shift in the
#    propagation medium.
#    Such shifts can occur in corrlation traces in case of a drifting clock.
#    This function ``applies`` the stretches. To correct for stretching
#    estimated with :class:`~miic.core.stretch_mod.time_stretch_estimate`you
#    need to apply negative stretching.
#    :type corr_data: :class:`~numpy.ndarray`
#    :param corr_data: 2d ndarray containing the correlation functions that are
#        to be shifted.
#        One for each row.
#    :type stretch: :class:`~numpy.ndarray`
#    :param stretch: ndarray with stretch.shape[0] = corr_data.shape[0]
#        containing the stretches relative units.
#    :rtype: :class:`~numpy.ndarray`
#    :return: **stretched_mat**: stretched version of the input matrix
#    """
#
#    # Mat must be a 2d vector in every case so
#    mat = np.atleast_2d(corr_data)
#    # check input
#    # stretch is just a 1d array
#    if len(stretch.shape) == 1:
#        t_stretch = np.zeros([stretch.shape[0], 1])
#        t_stretch[:, 0] = stretch
#        stretch = t_stretch
#    # stretch has the wrong length
#    elif stretch.shape[0] != mat.shape[0]:
#        msg = 'InputError: shift.shape[0] must be equal corr_data.shape[0]'
#        log.warning(msg)
#        return 0
#    # shift has multiple columns (multiple measurements for the same time)
#    if stretch.shape[1] > 1:
#        stretch = np.delete(stretch, np.arange(1, stretch.shape[1]), axis=1)
#
#    # taper and extend the reference trace to avoid interpolation
#    # artefacts at the ends of the trace
#    taper = cosine_taper(mat.shape[1], 0.05)
#    mat *= np.tile(taper, [mat.shape[0], 1])
#
#    # time axis
#    if single_sided:
#        time_idx = np.arange(mat.shape[1])
#    else:
#        time_idx = np.arange(mat.shape[1]) - (mat.shape[1] - 1.) / 2.
#
#    # allocate space for the result
#    stretched_mat = np.zeros_like(mat)
#
#    # stretch every line
#    for (ii, line) in enumerate(mat):
#        s = UnivariateSpline(time_idx, line, s=0)
#        stretched_mat[ii, :] = s(time_idx * np.exp(-stretch[ii]))
#
#    return stretched_mat
#
#
#if BC_UI:
#    class _time_stretch_apply_view(HasTraits):
#        shift = List()
#        corr_data = List()
#        trait_view = View(Item('shift'), Item('corr_data'))
#
#
#def multi_ref_creation(corr_mat,
#                       rtime,
#                       freq=30,
#                       use_break_point=False,
#                       break_point=None):
#    """ Create the multi-reference traces
#    This function creates multi-reference traces according with the given
#    frequency.
#    In case of a break-point is passed, the intervals to consider for the
#    reference traces creation are symmetric respect to it.
#    :type corr_mat: :class:`~numpy.ndarray`
#    :param corr_mat: Correlation matrix with one correlation function on each
#        row
#    :type rtime: :class:`~numpy.array` of :class:`~datetime.datetime` objects
#    :param rtime: Time vector associated to the given correlation matrix. Its
#        lenght must be equal to the numer of columns of the `corr_mat`
#        parameter
#    :type freq: int
#    :param freq: One reference trace every `freq` days
#    :type use_break_point: bool
#    :param use_break_point: If `True` the reference traces are calculated in
#        different intervals of `freq` days symmetric respect to the
#        `break_point`
#    :type break_point: string
#    :param break_point: Brake point expressed as "YYYY-MM-DD"
#    """
#
#    from pandas import DataFrame
#    # do a local import here to ensure compatibility with older pandas
#    # versions outside this function
#    from pandas import date_range
#
#    if use_break_point and break_point == None:
#        log.error("A break point must be passed")
#        return None
#
#    if use_break_point and type(break_point) != str:
#        log.warning("wrong break_point format")
#        return None
#
#    if use_break_point:
#        bp = from_str_to_datetime(break_point, datetimefmt=True)
#
#        f_frw = "%iD" % int(freq)
#        f_bck = "-%iD" % int(freq)
#
#        # Time "backward" intervals starting from the break_point and going
#        # back to the first day available in the data.
#        dr = date_range(rtime[rtime <= bp].max(),
#                        rtime[rtime < bp].min(),
#                        freq=f_bck)[::-1]
#
#        # Time "forward" intervals starting from the break_point and going
#        # ahead to the last day available in the data.
#        dr1 = date_range(bp, rtime.max(), freq=f_frw)[1:]  # break_point must
#                                                            # be removed
#
#        dr = dr.append(dr1)
#    else:
#        f = "%iD" % int(freq)
#        dr = date_range(np.min(rtime), np.max(rtime), freq=f)
#
#
#    # DataFrame creation
#    df = DataFrame(corr_mat, index=rtime)
#
#    # GroupBy the given intervals
#    dfg = df.groupby(lambda x: np.sum(np.ones(dr.shape)[dr <= x]))
#
#    # Calculate the reference traces averaging the traces on each interval
#    df_ref = dfg.mean()
#
#    # Take the raw output
#    ref_mat = df_ref.values
#
#    return ref_mat
#
#
#def multi_ref_vchange_and_align(corr_data,
#                      ref_trs,
#                      tw=None,
#                      stretch_range=0.1,
#                      stretch_steps=100,
#                      sides='both',
#                      return_sim_mat=False,
#                      remove_nans=True):
#    """ Multi-reference dv estimate and alignment
#    :type corr_data: :class:`~numpy.ndarray`
#    :param corr_data: 2d ndarray containing the correlation functions.
#        One for each row.
#    :type ref_trc: :class:`~numpy.ndarray`
#    :param ref_trc: 1D array containing the reference trace to be shifted and
#        compared to the individual traces in ``mat``
#    :type tw: list of :class:`~numpy.ndarray` of int
#    :param tw: list of 1D ndarrays holding the indices of sampels in the time
#        windows to be use in the time shift estimate. The sampels are counted
#        from the zero lag time with the index of the first sample being 0. If
#        ``tw = None`` the full time range is used.
#    :type stretch_range: scalar
#    :param stretch_range: Maximum amount of relative stretching.
#        Stretching and compression is tested from ``-stretch_range`` to
#        ``stretch_range``.
#    :type stretch_steps: scalar`
#    :param stretch_steps: Number of shifted version to be tested. The
#        increment will be ``(2 * stretch_range) / stretch_steps``
#    :type single_sided: bool
#    :param single_sided: if True zero lag time is on the first sample. If
#        False the zero lag time is in the center of the traces.
#    :type sides: str
#    :param sides: Side of the reference matrix to be used for the stretching
#        estimate ('both' | 'left' | 'right' | 'single') ``single`` is used for
#        one-sided signals from active sources with zero lag time is on the
#        first sample. Other options assume that the zero lag time is in the
#        center of the traces.
#    :type return_sim_mat: bool
#    :param return_sim_mat: If `True` the returning dictionary contains also the
#        similarity matrix `sim_mat'.
#    :type remove_nans: bool
#    :param remove_nans: If `True` applay :func:`~numpy.nan_to_num` to the
#        given correlation matrix before any other operation.
#    :rtype: Dictionary
#    :return: **dv**: Dictionary with the following keys
#        *corr*: 2d ndarray containing the correlation value for the best
#            match for each row of ``mat`` and for each time window.
#            Its dimension is: :func:(len(tw),mat.shape[1])
#        *value*: 2d ndarray containing the stretch amount corresponding to
#            the best match for each row of ``mat`` and for each time window.
#            Stretch is a relative value corresponding to the negative relative
#            velocity change -dv/v.
#            Its dimension is: :func:(len(tw),mat.shape[1])
#        *sim_mat*: 3d ndarray containing the similarity matricies that
#            indicate the correlation coefficient with the reference for the
#            different time windows, different times and different amount of
#            stretching.
#            Its dimension is: :py:func:`(len(tw),mat.shape[1],len(strvec))`
#        *second_axis*: It contains the stretch vector used for the velocity
#            change estimate.
#        *vale_type*: It is equal to 'stretch' and specify the content of
#            the returned 'value'.
#        *method*: It is equal to 'single_ref' and specify in which "way" the
#            values have been obtained.
#    """
#
#    # FIX: Trace back this problem to the original source and remove
#    # this statement
#    if remove_nans:
#        corr_data = np.nan_to_num(corr_data)
#        ref_trs = np.nan_to_num(ref_trs)
#
#    if tw and len(tw) > 1:
#        log.warning("The multi-reference vchange evaluation doesn't handle "
#                    "multiple time windows. Only the first time-window will "
#                    "be used")
#        tw = tw[0]
#
#    multi_ref_panel = multi_ref_vchange(corr_data,
#                                        ref_trs,
#                                        tw=tw,
#                                        stretch_range=stretch_range,
#                                        stretch_steps=stretch_steps,
#                                        sides=sides,
#                                        remove_nans=remove_nans)
#
#    n_ref = len(multi_ref_panel.keys())
#
#    if n_ref > 1:
#        dv = estimate_reftr_shifts_from_dt_corr(multi_ref_panel,
#                                                return_sim_mat=return_sim_mat)
#    else:
#        # changed key here
#        dv = multi_ref_panel['reftr_0']
#
#    return dv
#
#
#if BC_UI:
#    class _multi_ref_vchange_and_align_view(HasTraits):
#
#        stretch_range = Float(0.01)
#        stretch_steps = Int(100)
#        sides = Enum('both', 'left', 'right', 'sinlge')
#        return_sim_mat = Bool(False)
#
#        trait_view = View(Item('stretch_range'),
#                          Item('stretch_steps'),
#                          Item('sides'),
#                          Item('return_sim_mat'))
#
#
#def multi_ref_vchange(corr_data,
#                      ref_trs,
#                      tw=None,
#                      stretch_range=0.1,
#                      stretch_steps=100,
#                      sides='both',
#                      remove_nans=True):
#    """ Velocity change estimate with single or multiple reference traces.
#    This function estimates the velocity change corresponding to each row of
#    the ``corr_data`` matrix respect to the reference trace/s passed in
#    ``ref_trs``.
#    The velocity change is estimated comparing each correlation function stored
#    in the ``corr_data`` matrix (one for each row) with ``stretch_steps``
#    stretched versions of reference/s trace stored in ``ref_trs``.
#    The maximum amount of stretching may be passed in ``stretch_range``.
#    The best match (stretching amount and corresponding correlation value) is
#    calculated on different time windows. If ``tw = None`` the stretching is
#    estimated on the whole trace.
#    The output is a dictionary with keys of the form ``"reftr_%d" % i``: One
#    for each reference trace. The corresponding ``value`` is aslo a dictionary
#    that has a structure conforming with
#    :py:class:`~miic.core.stretch_mod.time_stretch_estimate` output.
#    :type corr_data: :class:`~numpy.ndarray`
#    :param corr_data: 2d ndarray containing the correlation functions.
#        One for each row.
#    :type ref_trc: :class:`~numpy.ndarray`
#    :param ref_trc: 1D array containing the reference trace to be shifted and
#        compared to the individual traces in ``mat``
#    :type tw: list of :class:`~numpy.ndarray` of int
#    :param tw: list of 1D ndarrays holding the indices of sampels in the time
#        windows to be use in the time shift estimate. The sampels are counted
#        from the zero lag time with the index of the first sample being 0. If
#        ``tw = None`` the full time range is used.
#    :type stretch_range: scalar
#    :param stretch_range: Maximum amount of relative stretching.
#        Stretching and compression is tested from ``-stretch_range`` to
#        ``stretch_range``.
#    :type stretch_steps: scalar`
#    :param stretch_steps: Number of shifted version to be tested. The
#        increment will be ``(2 * stretch_range) / stretch_steps``
#    :type single_sided: bool
#    :param single_sided: if True zero lag time is on the first sample. If
#        False the zero lag time is in the center of the traces.
#    :type sides: str
#    :param sides: Side of the reference matrix to be used for the stretching
#        estimate ('both' | 'left' | 'right' | 'single') ``single`` is used for
#        one-sided signals from active sources with zero lag time is on the
#        first sample. Other options assume that the zero lag time is in the
#        center of the traces.
#    :type remove_nans: bool
#    :param remove_nans: If `True` applay :func:`~numpy.nan_to_num` to the
#        given correlation matrix before any other operation.
#    :rtype: dictionary
#    :return: **multi_ref_panel**: It is a dictionary that contains as much
#        dictionaries as reference traces have been used. The key format is
#        ``"reftr_%d" % i`` and the corresponding value is also a dictionary
#        with the structure described in
#        :py:class:`~miic.core.stretch_mod.time_stretch_estimate`
#    """
#
#    # FIX: Trace back this problem to the original source and remove
#    # this statement
#    if remove_nans:
#        corr_data = np.nan_to_num(corr_data)
#        ref_trs = np.nan_to_num(ref_trs)
#
#    # remove 1-dimensions
#    ref_trs = np.squeeze(ref_trs)
#
#    # check how many reference traces have been passed
#    try:
#        reftr_count, _ = ref_trs.shape
#    except ValueError:  # An array is passed
#        reftr_count = 1
#
#    # Distionary that will hold all the results
#    multi_ref_panel = {}
#
#    # When there is just 1 reference trace no loop is necessary
#    if reftr_count == 1:
#        key = "reftr_0"
#        value = time_stretch_estimate(corr_data,
#                                      ref_trc=ref_trs,
#                                      tw=tw,
#                                      stretch_range=stretch_range,
#                                      stretch_steps=stretch_steps,
#                                      sides=sides,
#                                      remove_nans=remove_nans)
#        multi_ref_panel.update({key: value})
#    else:  # For multiple-traces loops
#        for i in range(reftr_count):
#            ref_trc = ref_trs[i]
#            key = "reftr_%d" % int(i)
#            value = time_stretch_estimate(corr_data,
#                                          ref_trc=ref_trc,
#                                          tw=tw,
#                                          stretch_range=stretch_range,
#                                          stretch_steps=stretch_steps,
#                                          sides=sides,
#                                          remove_nans=remove_nans)
#            multi_ref_panel.update({key: value})
#
#    return multi_ref_panel
#
#
#if BC_UI:
#    class _multi_ref_vchange_view(HasTraits):
#
#        stretch_range = Float(0.01)
#        stretch_steps = Int(100)
#        sides = Enum('both', 'left', 'right', 'sinlge')
#
#        trait_view = View(Item('stretch_range'),
#                          Item('stretch_steps'),
#                          Item('sides'))
#
#
#def est_shift_from_dt_corr(dt1, dt2, corr1, corr2):
#    """ Estimation of a baseline shift between velocity-change measurements
#    preformed with different references.
#    The use of different reference traces obtaind from different reference
#    periods will result in a shift of the velocity-change curves that ideally
#    characterizes the velocity variation between the two reference periods.
#    Instead of directly measuring this velocity change from the two reference
#    traces it is calulated here as the weighted average of the point
#    differences between the two velocity-change curves weighted by their
#    inverse variance according to Weaver et al. GJI 2011 (On the precision of
#    noise correlation interferometry)
#    Input vertors must all be of the same lenth.
#    :type dt1: :class:`~numpy.ndarray`
#    :pram dt1: Velocity variation measured for reference A
#    :type dt2: :class:`~numpy.ndarray`
#    :pram dt2: Velocity variation measured for reference B
#    :type corr1: :class:`~numpy.ndarray`
#    :pram corr1: Correlation between velocity corrected trace and reference A
#    :type corr2: :class:`~numpy.ndarray`
#    :pram corr2: Correlation between velocity corrected trace and reference B
#    """
#
#    # Create maked arrays so that NaNs are handled properly
#    dt1 = np.ma.masked_array(dt1, mask=(np.isnan(dt1) | np.isinf(dt1)),
#                             fill_value=0)
#    dt2 = np.ma.masked_array(dt2, mask=(np.isnan(dt2) | np.isinf(dt2)),
#                             fill_value=0)
#
#    corr1 = np.ma.masked_array(corr1, mask=(np.isnan(corr1) | np.isinf(corr1)),
#                               fill_value=0)
#    corr2 = np.ma.masked_array(corr2, mask=(np.isnan(corr2) | np.isinf(corr2)),
#                               fill_value=0)
#
#    # Remove the points where the correlation is 0
#    no_zero = (corr1 > 0) & (corr2 > 0)
#    corr1 = corr1[no_zero]
#    corr2 = corr2[no_zero]
#
#    # Estimate the point-variance for the two curves
#    var1 = (1 - corr1 ** 2) / (4 * corr1 ** 2)
#    var2 = (1 - corr2 ** 2) / (4 * corr2 ** 2)
#
#    # Calculate the point-weight
#    wgt = 1 / (var1 + var2)
#    mask = (corr1 > 0.999) & (corr2 > 0.999)
#    wgt = wgt[~mask]
#
#    # This saves from returning a masked value
#    try:
#        wgt_s = np.sum(wgt).filled(np.nan)
#    except:
#        wgt_s = np.sum(wgt)
#
#    # Calculate the shift and the total weight as a cumulative sum of the
#    # weighted average of the two curves
#    shift = np.sum((dt1[~mask] - dt2[~mask]) * wgt) / wgt_s
#    comb_corr = np.sum((corr1[~mask] + corr2[~mask]) * wgt) / wgt_s
#
#    # This saves from returning masked values
#    try:
#        shift = shift.filled(np.nan)
#    except:
#        pass
#
#    try:
#        comb_corr = comb_corr.filled(np.nan)
#    except:
#        pass
#
#    return comb_corr, shift
#
#
#def estimate_reftr_shifts_from_dt_corr(multi_ref_panel, return_sim_mat=False):
#    """ Combine velocity-change measurements of the same data performed with
#    different references to a single curve.
#    For a set of velocity-change measurements performed with different
#    references this function estimates the relative offsets between all pairs
#    of the measurements as a weighted average of their difference with the
#    function :py:class:`~miic.core.stretch_mod.est_shift_from_dt_corr`.
#    A least squares solution in computed that combines the pairwise
#    differences to a consistent set of reference shifts. These shifts should
#    be similar to the velocity variations measured between the reference
#    traces. The consistent set of reference shifts is used to correct i.e.
#    shift the similarity matricies to a common reference. Finally the
#    corrected similarity matrices are averaged resulting in a single matrix
#    that is interpreted as before. The position of the crest is the combined
#    velocity change and the height of the crest is the correlation value.
#    :type multi_ref_panel: dictionay
#    :param multi_ref_panel: It is a dictionary with one (key,value) pair
#        for each reference trace. Its structure is described
#        in :py:class:`~miic.core.stretch_mod.multi_ref_vchange`
#    :type return_sim_mat: bool
#    :param return_sim_mat: If `True` the returning dictionary contains also the
#        similarity matrix `sim_mat'.
#    :rtype: Dictionary
#    :return: **dv**: Dictionary with the following keys
#        *corr*: 2d ndarray containing the correlation value for the best
#            match for each row of ``mat`` and for each time window.
#            Its dimension is: :func:(len(tw),mat.shape[1])
#        *value*: 2d ndarray containing the stretch amount corresponding to
#            the best match for each row of ``mat`` and for each time window.
#            Stretch is a relative value corresponding to the negative relative
#            velocity change -dv/v.
#            Its dimension is: :func:(len(tw),mat.shape[1])
#        *sim_mat*: 3d ndarray containing the similarity matricies that
#            indicate the correlation coefficient with the reference for the
#            different time windows, different times and different amount of
#            stretching.
#            Its dimension is: :py:func:`(len(tw),mat.shape[1],len(strvec))`
#        *second_axis*: It contains the stretch vector used for the velocity
#            change estimate.
#        *vale_type*: It is equal to 'stretch' and specify the content of
#            the returned 'value'.
#        *method*: It is equal to 'single_ref' and specify in which "way" the
#            values have been obtained.
#    """
#
#    # Vector with the stretching amount
#    stretch_vect = multi_ref_panel['reftr_0']['second_axis']
#    delta = stretch_vect[1] - stretch_vect[0]
#
#    n_ref = len(multi_ref_panel.keys())
#
#    corr = []
#    shift = []
#
#    if n_ref > 1:
#
#        # Loop over reftr
#        for reftr1 in np.sort(multi_ref_panel.keys()):
#            ref_idx = [reftr for reftr in np.sort(multi_ref_panel.keys())
#                       if reftr != reftr1]
#            for reftr2 in ref_idx:
#                ccorr, sshift = est_shift_from_dt_corr(
#                            np.squeeze(multi_ref_panel[reftr1]['value']),
#                            np.squeeze(multi_ref_panel[reftr2]['value']),
#                            np.squeeze(multi_ref_panel[reftr1]['corr']),
#                            np.squeeze(multi_ref_panel[reftr2]['corr']))
#                corr.append(ccorr)
#                shift.append(sshift)
#
#        G = _create_G(len(multi_ref_panel.keys()))
#        W = np.diag(np.array(corr))
#        D = np.array(shift)
#
#        # This conversion is necessary to achive the same speed with diffrent
#        # .dot implementation on different numpy versions.
#        D = np.nan_to_num(D)
#        W = np.nan_to_num(W)
#
#        GTW = np.dot(G.T, W)
#
#        # to_invert = np.dot(G.T, np.dot(W, G))
#        to_invert = np.dot(GTW, G)
#
#        # TODO: Get rid of the matrix inversion
#        left_op = np.linalg.pinv(to_invert)
#
#        # This conversion is necessary to achive the same speed with diffrent
#        # .dot implementation on different numpy versions.
#        left_op = np.nan_to_num(left_op)
#
#        m = np.dot(left_op, np.dot(GTW, D))
#        m = np.hstack((0, m))
#        m = m - np.mean(m)
#
#        # How many samples (int) each sim matrix must be rolled
#        m = np.around(m / delta, out=np.zeros_like(m, dtype='int32'))
#
#        row, col = np.squeeze(multi_ref_panel['reftr_0']['sim_mat']).shape
#
#        stmp = np.zeros((row, col, n_ref))
#
#        # Roll the sim_mat
#        for (i, reftr) in enumerate(np.sort(multi_ref_panel.keys())):
#
#            stmp[:, :, i] = \
#                np.roll(np.squeeze(multi_ref_panel[reftr]['sim_mat']),
#                        m[i],
#                        axis=1)
#
#        # Create a msked array to evaluate the mean
#        stmp_masked = np.ma.masked_array(stmp,
#                                         mask=(np.isnan(stmp) | \
#                                               np.isinf(stmp)),
#                                         fill_value=0)
#
#        # Evaluate the sim_mat for the multi-ref approach as the mean
#        # of the rolled sim_mat corresponfing to the individual reference
#        # traces
#        # When operating with masked arrays this operation creates a new
#        # object so the default fill_values returns to be 1e20
#        bsimmat = np.mean(stmp_masked, axis=2)
#
#        corr = np.max(bsimmat, axis=1)
#        dt = np.argmax(bsimmat, axis=1)
#
#        dt = stretch_vect[dt]
#
#        # Remove the mask
#        bsimmat = bsimmat.filled(np.nan)
#
#        try:
#            corr = corr.filled(np.nan)
#        except:
#            pass
#
#        try:
#            dt = dt.filled(np.nan).astype('int')
#        except:
#            pass
#
#        ret_dict = {'corr': corr,
#                    'value': dt,
#                    'second_axis': stretch_vect,
#                    'value_type': np.array(['stretch']),
#                    'method': np.array(['multi_ref'])}
#
#        if return_sim_mat:
#            ret_dict.update({'sim_mat': bsimmat})
#
#        return ret_dict
#
#    else:
#        log.error("For a single reference trace use the appropriate funtion")
#        return None


def time_windows_creation(starting_list, t_width):
    """ Time windows creation.
    A matrix containing one time window for each row is created. The starting
    samples of each one of them are passed in the ``starting_list`` parameter.
    The windows length ``t_width`` can be scalar or a list of values.
    In the latter case both lists ``starting_list`` and ``t_width`` must
    have the same length.
    :type starting_list: list
    :param starting_list: List of starting points
    :type t_width: int or list of int
    :param t_width: Windows length
    :rtype: :class:`~numpy.ndarray`
    :return: **tw_mat**: 2d ndarray containing the indexes of one time window
        for each row
    """

    if not np.isscalar(starting_list):
        if not np.isscalar(t_width) and len(t_width) != len(starting_list):
            raise ValueError("t_width must be a scalar or list of scalars of\
                            the same length as starting_list")

    tw_list = []

    if np.isscalar(starting_list):
        if np.isscalar(t_width):
            wlen = t_width
        else:
            wlen = t_width[0]
        tw_list.append(np.arange(starting_list, starting_list + wlen, 1))
    else:
        for (ii, cstart) in enumerate(starting_list):
            if np.isscalar(t_width):
                wlen = t_width
            else:
                wlen = t_width[ii]
            tw_list.append(np.arange(cstart, cstart + wlen, 1))

    return tw_list


#if BC_UI:
#    class _time_windows_creation_view(HasTraits):
#        starting_list = List(Int)
#        t_width = List(Int(50))
#
#        trait_view = View(Tabbed(Item('starting_list',
#                                      height=150,
#                                      width=80),
#                                 Item('t_width',
#                                      label='Win width in samples',
#                                      height=150,
#                                      width=80)))
#
#
#def _create_G(n_ref):
#    """ Create the G matrix for the multi-trace alignment
#    """
#
#    G = None
#    for jj in range(n_ref):
#        line = range(n_ref)
#        tline = [i for i in line if i != jj]
#        tG = np.zeros((len(tline), len(tline)))
#        if jj > 0:
#            tG[:, jj - 1] = -1
#        for ii in range(len(tline)):
#            if tline[ii] > 0:
#                tG[ii, tline[ii] - 1] = 1
#        if G is None:
#            G = tG
#        else:
#            G = np.vstack((G, tG))
#    return G
#
#
#def _id_from_stats(stats, ret_id_dict=False):
#    id_dict = {}
#    stats = flatten_recarray(stats)
#    keywords = ['network', 'station', 'location', 'channel']
#    tr_id = '.'.join([stats[key].replace('-', '') for key in \
#                      keywords if isinstance(stats[key], str)])
#    if not ret_id_dict:
#        return tr_id
#    for key in keywords:
#        if stats[key] != []:
#            id_dict[key] = stats[key]
#        else:
#            id_dict[key] = ''
#    return tr_id, id_dict
#
#
#def dv_panel_creation(dv_list, save_dir='./save'):
#    """ Create :class:`pandas.Panel` objects recombining corr and dv curves
#    This function creates two :py:class:`pandas.Panel` object recombining
#    respectively all correlation `curve` and `dv/v` curve created for
#    a specific dataset and one :py:class:`pandas.DataFrame` containing all
#    the metainformation in each `stats` object associated with one stations
#    pair.
#    The function accept a list of dv dictionaries that could be created, e.g.,
#    with the function
#    :py:func:`~miic.core.stretch_mod.velocity_change_estimate` or with
#    :py:func:`~miic.core.stretch_mod.multi_ref_vchange_and_align`
#    that are checked for a consistent structure through
#    :py:func:`~miic.core.miic_utils.dv_check`.
#    Patterns that are processed without errors are logged in a file
#    ``success_patterns.npy`` saved in the `save_dir` directory.
#    :type dv_list: list of dv dictionaries
#    :param dv_list: List of dv dictionaries
#    :type save_dir: string
#    :param save_dir: Where the two Panels and the one DataFrame will be saved
#    :rtype: dict
#    :return: **panel_dict**: dictionary that has this structure:
#        *dv*: Velocity change panel
#        *corr*: Correlation panel
#        *stats*: Meta-information stored in a DataFrame
#    """
#    from pandas import DataFrame, Series, Panel
#    stats_df = DataFrame(columns=['evel',
#                                  'endtime',
#                                  'dist',
#                                  'network',
#                                  'channel',
#                                  'stlo',
#                                  'baz',
#                                  'stel',
#                                  'evla',
#                                  'npts',
#                                  'station',
#                                  'location',
#                                  'starttime',
#                                  'sampling_rate',
#                                  'stla',
#                                  'az',
#                                  'evlo'])
#
#    base_patterns = []
#
#    for (i, dv_dict) in enumerate(dv_list):
#
#        check_state = dv_check(dv_dict)
#
#        # Check if the dv dictionary is "correct"
#        if check_state['is_incomplete']:
#            log.error("Incomplete dv. Possible errors:")
#            for key in check_state:
#                if key is not 'is_incomplete':
#                    log.error("%s: %s", key, check_state[key])
#            continue
#
#        # Extract the stats obj and use it to generate the pattern
#        stats = flatten_recarray(dv_dict['stats'])
#        pattern = _id_from_stats(stats)
#
#        # Add the current stats obj to the stats dataframe
#        cdf = DataFrame(stats, index=[pattern])
#        stats_df = stats_df.append(cdf, verify_integrity=True)
#
#        # Generate the time index
#        time = convert_time(dv_dict['time'])
#
#        if time is None:
#            log.error("Time format error: skip pattern %s", pattern)
#            continue
#
#        # Extract the corr and dv arrays
#        corr = np.squeeze(dv_dict['corr'])
#        try:
#            dv = np.squeeze(dv_dict['value'])  # New style dictionaries
#        except KeyError:  # Try the old style format
#            try:
#                dv = np.squeeze(dv_dict['dv'])
#            except KeyError:
#                try:
#                    dv = np.squeeze(dv_dict['dt'])
#                except Exception:
#                    raise Exception
#
#        # Adapt dv and corr in case they are single curves
#        try:
#            _, _ = dv.shape
#        except ValueError:
#            dv = dv[np.newaxis, :]
#            corr = corr[np.newaxis, :]
#
#        if i == 0:
#            xyz = {}
#            pqr = {}
#            win_lab = []
#            for k in range(len(dv)):
#                lab = 'win-%i' % k
#                win_lab.append(lab)
#                xyz[lab] = {}
#                pqr[lab] = {}
#
#        for j in range(len(dv)):
#            cc = Series(corr[j], index=time)
#            # ss = Series(dv[j] - 1, index=time)
#            ss = Series(dv[j], index=time)
#            xyz[win_lab[j]].update({pattern: ss})
#            pqr[win_lab[j]].update({pattern: cc})
#
#        base_patterns.append(pattern)
#
#    dv = Panel(xyz)
#    corr = Panel(pqr)
#
#    dv.save(os.path.join(save_dir, 'dv.pickle'))
#    corr.save(os.path.join(save_dir, 'corr.pickle'))
#    if len(stats_df) > 0:
#        stats_df.save(os.path.join(save_dir, 'stats.pickle'))
#
#    np.save(os.path.join(save_dir, 'success_patterns.npy'), \
#        base_patterns)
#
#    panel_dict = {}
#    panel_dict['dv'] = dv
#    panel_dict['corr'] = corr
#    if len(stats_df) > 0:
#        panel_dict['stats'] = stats_df
#
#    return panel_dict




# Copyright 2017-2019 Tom Eulenfeld, MIT license
"""Utility functions"""

from importlib import import_module
import logging
import numbers
import os
from pkg_resources import resource_filename
import shutil
import sys
import tempfile

import numpy as np
from obspy import UTCDateTime as UTC
import scipy.signal
import tqdm


log = logging.getLogger('yam.util')


class YamError(Exception):
    pass


class ParseError(YamError):
    pass


class ConfigError(YamError):
    pass


def _analyze_key(key):
    if key.startswith('/'):
        key = key[1:]
    if '/' in key:
        key = key.split('/', 1)[0]
    return ''.join([k[0] for k in key.split('_')])


def _get_fname(io, key):
    fname = (io['stretch'] if 't' in _analyze_key(key)
             else io['stack'] if 's' in _analyze_key(key)
             else io['corr'])
    return fname


def _corr_id(trace):
    st = trace.stats
    try:
        id_ = (st.network1, st.station1, st.location1, st.channel1,
               st.network2, st.station2, st.location2, st.channel2)
    except AttributeError:
        log.error('meta data error in read file, trace: %s, stats: %s',
                  trace, st)
        return
    return '%s.%s.%s.%s-%s.%s.%s.%s' % id_


def _filter(stream, filter):
    if filter[0] is None:
        stream.filter("lowpass", freq=filter[1])
    elif filter[1] is None:
        stream.filter("highpass", freq=filter[0])
    else:
        stream.filter("bandpass", freqmin=filter[0], freqmax=filter[1])


def _load_func(modulename, funcname):
    """Load and return function from Python module"""
    sys.path.append(os.path.curdir)
    module = import_module(modulename)
    sys.path.pop(-1)
    func = getattr(module, funcname)
    return func


def _seedid2meta(seedid):
    net, sta, loc, cha = seedid.split('.')
    smeta = {'network': net, 'station': sta, 'location': loc, 'channel': cha}
    return smeta


def _time2sec(time):
    """Convert string (e.g. 1d or 0.5h) to seconds"""
    if not isinstance(time, numbers.Number):
        time, unit = float(time[:-1]), time[-1]
        assert unit in 'dh'
        time *= 24 * 3600 if unit == 'd' else 3600
    return time


def _trim(tr, time_interval):
    """Trim traces of stream relative to mid sample"""
    starttime = tr.stats.starttime
    mid = starttime + (tr.stats.endtime - starttime) / 2
    if time_interval is not None:
        start, end = time_interval
        if start is not None:
            start = mid + start
        if end is not None:
            end = mid + end
        tr.trim(start, end)
    times = tr.times(reftime=mid)
    tr.stats.starttime = starttime
    return times


def _trim_time_period(stream, time_period):
    """Restrict traces of stream to given time period"""
    if time_period is None:
        return
    starttime, endtime = time_period
    traces = [tr for tr in stream if
              (starttime is None or tr.stats.starttime >= UTC(starttime)) and
              (endtime is None or tr.stats.starttime < UTC(endtime))]
    stream.traces = traces


def create_config(conf='conf.json', tutorial=False, less_data=False):
    """Create JSON config file and download tutorial data if requested"""
    shutil.copyfile(resource_filename('yam', 'conf_example.json'), conf)
    temp_dir = os.path.join(tempfile.gettempdir(), 'yam_example_data')
    template = os.path.join(temp_dir, 'example_data')
    station_template = os.path.join(temp_dir, 'example_inventory')
    try:
        num_files = (len([name for name in os.listdir(template)]),
                     len([name for name in os.listdir(station_template)]))
    except FileNotFoundError:
        num_files = (0, 0)
    if tutorial and (num_files[0] < (9 if less_data else 54) or
                     num_files[1] < 3):
        print('Download example data from Geofon')
        from obspy import UTCDateTime as UTC
        from obspy.clients.fdsn.mass_downloader import (
            GlobalDomain, Restrictions, MassDownloader)
        domain = GlobalDomain()
        restrictions = Restrictions(
            starttime=UTC('2010-02-04' if less_data else '2010-02-01'),
            endtime=UTC('2010-02-06' if less_data else '2010-02-15'),
            network='CX', station='PATCX', location=None,
            channel_priorities=["BH[ZN]"], chunklength_in_sec=86400,
            reject_channels_with_gaps=False, minimum_length=0.5)
        mdl = MassDownloader(providers=['GFZ'])
        kw = dict(threads_per_client=1, download_chunk_size_in_mb=200)
        mdl.download(domain, restrictions, template, station_template,  **kw)
        restrictions.station = 'PB06'
        if not less_data:
            restrictions.endtime = UTC('2010-02-12')
        mdl.download(domain, restrictions, template, station_template,  **kw)
        restrictions.station = 'PB01'
        restrictions.endtime = UTC('2010-02-04 08:00:00')
        restrictions.channel_priorities = ["BHZ"]
        mdl.download(domain, restrictions, template, station_template,  **kw)
        if not less_data:
            restrictions.starttime = UTC('2010-02-08 00:00:00')
            restrictions.endtime = UTC('2010-02-09 23:55:00')
            restrictions.channel_priorities = ["BHZ"]
            mdl.download(domain, restrictions, template, station_template,
                         **kw)
    if tutorial:
        dest_dir = os.path.dirname(conf)
        dest_dir_data = os.path.join(dest_dir, 'example_data')
        dest_dir_inv = os.path.join(dest_dir, 'example_inventory')
        if not os.path.exists(dest_dir_data):
            if less_data:
                ignore = shutil.ignore_patterns('*2010020[123]T000000Z__*',
                                                '*2010020[6-9]T000000Z__*',
                                                '*2010021?T000000Z__*')
            else:
                ignore = None
            shutil.copytree(template, dest_dir_data, ignore=ignore)
        if not os.path.exists(dest_dir_inv):
            shutil.copytree(station_template, dest_dir_inv)


# def _stream2matrix(stream):
#    """
#    Return array with data in time window (start, end) around relative.
#
#    'time' can stand for UTCDateTime, list of UTCDateTimes, header entry out of
#    ('ponset', 'sonset', 'startime', 'endtime') or 'middle'
#    :param stream: Stream object with data
#    :param trim: 2 time or float (seconds) relative to param=relative
#    :param relative: time, is needed if start or end in seconds (float)
#    :return: np.array of shape (N_stream, N_data)
#    """
#    if len(stream) == 0:
#        raise ValueError('Stream has length 0')
#    samp = [tr.stats.sampling_rate for tr in stream]
#    npts = [len(tr) for tr in stream]
#    if min(samp) != max(samp):
#        for tr in stream:
#            tr.decimate(int(tr.stats.sampling_rate) // min(samp))
#        log.warning('Downsampling stream because of differing sampling rate.')
#    if min(npts) != max(npts):
#        log.warning('Traces in stream have different NPTS. '
#                    'Difference: %d samples' % (max(npts) - min(npts)))
#    data = np.zeros((len(stream), max(npts)))
#    for i, trace in enumerate(stream):
#        data[i, :len(trace.data)] = trace.data
#    return data


def smooth(x, window_len=None, window='flat', method='zeros'):
    """Smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with the signal.

    :param x: the input signal (numpy array)
    :param window_len: the dimension of the smoothing window; should be an
        odd integer
    :param window: the type of window from 'flat', 'hanning', 'hamming',
        'bartlett', 'blackman'
        flat window will produce a moving average smoothing.
    :param method: handling of border effects\n
        'zeros': zero padding on both ends (len(smooth(x)) = len(x))\n
        'reflect': pad reflected signal on both ends (same)\n
        'clip': pad signal on both ends with the last valid value (same)\n
        None: no handling of border effects
        (len(smooth(x)) = len(x) - len(window_len) + 1)
    """
    if window_len is None:
        return x
    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")
    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")
    if window_len < 3:
        return x
    if window not in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is one of 'flat', 'hanning', 'hamming',"
                         "'bartlett', 'blackman'")
    if method == 'zeros':
        s = np.r_[np.zeros((window_len - 1) // 2), x,
                  np.zeros(window_len // 2)]
    elif method == 'reflect':
        s = np.r_[x[(window_len - 1) // 2:0:-1], x,
                  x[-1:-(window_len + 1) // 2:-1]]
    elif method == 'clip':
        s = np.r_[x[0] * np.ones((window_len - 1) // 2), x,
                  x[-1] * np.ones(window_len // 2)]
    else:
        s = x
    if window == 'flat':
        w = np.ones(window_len, 'd')
    else:
        w = getattr(np, window)(window_len)
    return scipy.signal.fftconvolve(w / w.sum(), s, mode='valid')


class IterTime():
    """Iterator yielding UTCDateTime objects between start- and endtime"""

    def __init__(self, startdate, enddate, dt=24 * 3600):
        self.startdate = startdate
        self.enddate = enddate
        self.dt = dt

    def __len__(self):
        return int((self.enddate - self.startdate) / self.dt)

    def __iter__(self):
        t = self.startdate
        while t <= self.enddate:
            yield t
            t += self.dt


# https://stackoverflow.com/a/38739634
# not working yet for parallel processing
class TqdmLoggingHandler(logging.Handler):
    def __init__(self, level=logging.NOTSET):
        super(self.__class__, self).__init__(level)

    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.tqdm.write(msg)
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)


LOGLEVELS = {0: 'CRITICAL', 1: 'WARNING', 2: 'INFO', 3: 'DEBUG'}


LOGGING_DEFAULT_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'capture_warnings': True,
    'formatters': {
        'file': {
            'format': ('%(asctime)s %(module)-6s%(process)-6d%(levelname)-8s'
                       '%(message)s')
        },
        'console': {
            'format': '%(levelname)-8s%(message)s'
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'console',
            'level': None,
        },
        'console_tqdm': {
            'class': 'yam.util.TqdmLoggingHandler',
            'formatter': 'console',
            'level': None,
        },
        'file': {
            'class': 'logging.FileHandler',
            'formatter': 'file',
            'level': None,
            'filename': None,
        },
    },
    'loggers': {
        'yam': {
            'handlers': ['console_tqdm', 'file'],
            'level': 'DEBUG',
            'propagate': False,
        },
        'py.warnings': {
            'handlers': ['console_tqdm', 'file'],
            'level': 'DEBUG',
            'propagate': False,
        }

    }
}



# Copyright 2017-2019 Tom Eulenfeld, MIT license
"""
Command line interface and main entry point
"""

import argparse
from argparse import SUPPRESS
from copy import deepcopy
import glob
import json
import logging
import logging.config
import sys
import time

import obspy

import yam.commands
from yam.util import (_load_func, create_config,
                      LOGLEVELS, LOGGING_DEFAULT_CONFIG, ParseError,
                      ConfigError)


log = logging.getLogger('yam')
log.addHandler(logging.NullHandler())


class ConfigJSONDecoder(json.JSONDecoder):
    def decode(self, s):
        """Decode JSON config with comments stripped"""
        s = '\n'.join(l.split('#', 1)[0] for l in s.split('\n'))
        return super(ConfigJSONDecoder, self).decode(s)


def configure_logging(loggingc, verbose=0, loglevel=3, logfile=None):
    if loggingc is None:
        loggingc = deepcopy(LOGGING_DEFAULT_CONFIG)
        if verbose > 3:
            verbose = 3
        loggingc['handlers']['console']['level'] = LOGLEVELS[verbose]
        loggingc['handlers']['console_tqdm']['level'] = LOGLEVELS[verbose]
        if logfile is None or loglevel == 0:
            del loggingc['handlers']['file']
            loggingc['loggers']['yam']['handlers'] = ['console_tqdm']
            loggingc['loggers']['py.warnings']['handlers'] = ['console_tqdm']
        else:
            loggingc['handlers']['file']['level'] = LOGLEVELS[loglevel]
            loggingc['handlers']['file']['filename'] = logfile
    logging.config.dictConfig(loggingc)
    logging.captureWarnings(loggingc.get('capture_warnings', False))


def __get_station(seedid):
    """Station name from seed id"""
    st = seedid.rsplit('.', 2)[0]
    if st.startswith('.'):
        st = st[1:]
    return st


def load_inventory(inventory):
    try:
        if not isinstance(inventory, obspy.Inventory):
            if isinstance(inventory, str):
                format_ = None
            else:
                inventory, format_ = inventory
            expr = inventory
            inventory = None
            for fname in glob.glob(expr):
                inv2 = obspy.read_inventory(fname, format_)
                if inventory is None:
                    inventory = inv2
                else:
                    inventory += inv2
            channels = inventory.get_contents()['channels']
            stations = list(set(__get_station(ch) for ch in channels))
            log.info('read inventory with %d stations', len(stations))
    except:
        log.exception('cannot read stations')
        return
    return inventory


def _get_kwargs(kwargs, id_):
    kw = kwargs[id_]
    if 'based_on' in kw:
        kw2 = kwargs[kw.pop('based_on')]
        kw2.update(kw)
        kw = kw2
    return kw


def run(command, conf=None, tutorial=False, less_data=False, pdb=False,
        **args):
    """Main entry point for a direct call from Python

    Example usage:

    >>> from yam import run
    >>> run(conf='conf.json')

    :param command: if ``'create'`` the example configuration is created,
       optionally the tutorial data files are downloaded

    For all other commands this function loads the configuration
    and construct the arguments which are passed to `run2()`

    All args correspond to the respective command line and
    configuration options.
    See the example configuration file for help and possible arguments.
    Options in args can overwrite the configuration from the file.
    E.g. ``run(conf='conf.json', bla='bla')`` will set bla configuration
    value to ``'bla'``.
    """
    if pdb:
        import traceback, pdb

        def info(type, value, tb):
            traceback.print_exception(type, value, tb)
            print
            # ...then start the debugger in post-mortem mode.
            pdb.pm()

        sys.excepthook = info
    if conf in ('None', 'none', 'null', ''):
        conf = None
    # Copy example files if create_config or tutorial
    if command == 'create':
        if conf is None:
            conf = 'conf.json'
        create_config(conf, tutorial=tutorial, less_data=less_data)
        return
    # Parse config file
    if conf:
        try:
            with open(conf) as f:
                conf = json.load(f, cls=ConfigJSONDecoder)
        except ValueError as ex:
            msg = 'Error while parsing the configuration: %s' % ex
            raise ConfigError(msg)
        except IOError as ex:
            raise ConfigError(ex)
        # Populate args with conf, but prefer args
        conf.update(args)
        args = conf
    run2(command, **args)


def run2(command, io,
         logging=None, verbose=0, loglevel=3, logfile=None,
         key=None, keys=None, corrid=None, stackid=None, stretchid=None,
         correlate=None, stack=None, stretch=None,
         **args):
    """
    Second main function for unpacking arguments

    Initialize logging, load inventory if necessary, load options from
    configuration dictionary into args
    (for correlate, stack and stretch commands) and run the corresponding
    command in `~yam.commands` module. If ``"based_on"`` key is set the
    configuration dictionary will be preloaded with the specified
    configuration.

    :param command: specified subcommand, will call one of
        `~yam.commands.start_correlate()`,
        `~yam.commands.start_stack()`,
        `~yam.commands.start_stretch()`,
        `~yam.commands.info()`,
        `~yam.commands.load()`,
        `~yam.commands.plot()`,
        `~yam.commands.remove()`
    :param logging,verbose,loglevel,logfile: logging configuration
    :param key: the key to work with
    :param keys: keys to remove (only remove command)
    :param correlate,stack,stretch: corresponding configuration dictionaries
    :param \*id: the configuration id to load from the config dictionaries
    :param \*\*args: all other arguments are passed to next called function
    """
    time_start = time.time()
    # Configure logging
    if command in ('correlate', 'stack', 'stretch'):
        configure_logging(loggingc=logging, verbose=verbose,
                          loglevel=loglevel, logfile=logfile)
    log.info('Yam version %s', yam.__version__)
    if key is not None and '/' in key:
        key, subkey = key.split('/', 1)
        subkey = '/' + subkey
    else:
        subkey = ''
    # load data plugin, discard unnecessary io kwargs
    data_plugin = io.get('data_plugin')
    if command == 'correlate' or (
            command in ('print', 'load', 'plot') and
            key in ('data', 'prepdata')):
        if data_plugin is not None:
            modulename, funcname = data_plugin.split(':')
            io['data'] = _load_func(modulename.strip(), funcname.strip())
    # pop plotting options
    if command != 'plot':
        for k in list(args.keys()):
            if k.startswith('plot_'):
                args.pop(k)
    # Start main routine
    if command == 'correlate':
        kw = _get_kwargs(correlate, corrid)
        kw['outkey'] = 'c' + corrid
        args.update(kw)
        io['inventory'] = load_inventory(kw.get('inventory') or
                                         io.get('inventory'))
        yam.commands.start_correlate(io, **args)
    elif command == 'stack':
        if stack is not None and stackid in stack:
            kw = _get_kwargs(stack, stackid)
        else:
            kw = {}
            if 'm' in stackid:
                kw['length'], kw['move'] = stackid.split('m')
            elif stackid in ('', 'all', 'None', 'none', 'null'):
                stackid = ''
                kw['length'] = None
            else:
                kw['length'] = stackid
        kw['outkey'] = key + '_s' + stackid
        args.update(kw)
        yam.commands.start_stack(io, key, subkey=subkey, **args)
    elif command == 'stretch':
        kw = _get_kwargs(stretch, stretchid)
        kw['outkey'] = key + '_t' + stretchid
        args.update(kw)
        yam.commands.start_stretch(io, key, subkey=subkey, **args)
    elif command == 'remove':
        yam.commands.remove(io, keys)
    elif command in ('info', 'print', 'load', 'plot', 'export'):
        if key == 'stations' or command == 'info' and key is None:
            io['inventory'] = load_inventory(io['inventory'])
        if key == 'prepdata':
            if corrid is None:
                msg = 'seed id, day and corrid need to be set for prepdata'
                raise ParseError(msg)
            kw = _get_kwargs(correlate, corrid)
            io['inventory'] = load_inventory(kw.get('inventory') or
                                             io['inventory'])
            args['prep_kw'] = kw
        if command == 'info':
            yam.commands.info(io, key=key, subkey=subkey,
                              config=(correlate, stack, stretch), **args)
        elif command == 'print':
            yam.commands.load(io, key=key + subkey, do='print', **args)
        elif command == 'load':
            yam.commands.load(io, key=key + subkey, do='load', **args)
        elif command == 'export':
            yam.commands.load(io, key=key + subkey, do='export', **args)
        else:
            yam.commands.plot(io, key + subkey, corrid=corrid, **args)
    elif command == 'scan':
        data_glob = yam.commands._get_data_glob(io['data'])
        print('Suggested call to obspy-scan:')
        if io.get('data_format') is not None:
            print('obspy-scan -f %s %s' % (io['data_format'], data_glob))
        else:
            print('obspy-scan %s' % data_glob)
        print('This will probably not work if you use the data_plugin '
              'configuration. '
              'For more options check obspy-scan -h.')
    else:
        raise ValueError('Unknown command')
    time_end = time.time()
    log.debug('used time: %.1fs', time_end - time_start)


def run_cmdline(args=None):
    """Main entry point from the command line"""
    # Define command line arguments
    from yam import __version__
    msg = ('Yam: Yet another monitoring tool using cross-correlations of '
           'ambient noise')
    epilog = 'To get help on a subcommand run: yam command -h'
    p = argparse.ArgumentParser(description=msg, epilog=epilog)
    version = '%(prog)s ' + __version__
    p.add_argument('--version', action='version', version=version)
    msg = 'configuration file to load (default: conf.json)'
    p.add_argument('-c', '--conf', default='conf.json', help=msg)
    msg = 'if an exception occurs start the debugger'
    p.add_argument('--pdb', action='store_true', help=msg)

    sub = p.add_subparsers(title='commands', dest='command')
    sub.required = True
    msg = 'create config file in current directory'
    p_create = sub.add_parser('create', help=msg)
    msg = 'preprocess and correlate data'
    p_correlate = sub.add_parser('correlate', help=msg)
    msg = 'stack correlations or stacked correlations'
    p_stack = sub.add_parser('stack', help=msg)
    msg = 'stretch correlations or stacked correlations'
    p_stretch = sub.add_parser('stretch', help=msg)
    msg = 'remove keys from HDF5 files'
    p_remove = sub.add_parser('remove', help=msg)
    msg = 'print information about project or objects'
    p_info = sub.add_parser('info', help=msg)
    msg = 'print objects'
    p_print = sub.add_parser('print', help=msg)
    msg = 'print suggested call for obspy-scan scipt (data availability)'
    p_scan = sub.add_parser('scan', help=msg)
    msg = 'plot objects'
    p_plot = sub.add_parser('plot', help=msg)
    msg = 'load objects into IPython session'
    p_load = sub.add_parser('load', help=msg)
    msg = ('export corrrelations or stacks to another file format '
           'for processing with other programs, '
           'preprocessed data can also be saved')
    p_export = sub.add_parser('export', help=msg)

    msg = 'additionaly create example files for tutorial'
    p_create.add_argument('--tutorial', help=msg, action='store_true')
    p_create.add_argument('--less-data', help=SUPPRESS, action='store_true')

    msg = 'use the data defined by this key (processing chain)'
    p_stack.add_argument('key', help=msg)
    p_stretch.add_argument('key', help=msg)
    msg = 'keys of data to be removed'
    p_remove.add_argument('keys', help=msg, nargs='+')
    msg = 'configuration key for correlation'
    p_correlate.add_argument('corrid', help=msg)
    msg = 'configuration key or explicit configuration for stacking'
    p_stack.add_argument('stackid', help=msg)
    msg = 'configuration key for stretching'
    p_stretch.add_argument('stretchid', help=msg)
    msg = 'key of reference trace (stack)'
    p_stretch.add_argument('--reftrid', help=msg)

    msg1 = 'more detailed information about '
    msg2 = ('stations, data or specific processing key '
            '(e.g. stations, data, '
            'c1_s1d, c1_s1d_t1/CX.PB06-CX.PB06/.BHZ-.BHZ)')
    p_info.add_argument('key', help=msg1 + msg2, nargs='?')
    msg1 = 'print contents of '
    msg2 = ('stations, data, preprocessed data or specific processing key '
            '(e.g. stations, data, prepdata, '
            'c1_s1d, c1_s1d_t1/CX.PB06-CX.PB06/.BHZ-.BHZ)')
    p_print.add_argument('key', help=msg1 + msg2)
    msg1 = 'load IPython session with contents of '
    p_load.add_argument('key', help=msg1 + msg2)
    msg1 = 'plot '
    p_plot.add_argument('key', help=msg1 + msg2)
    msg1 = 'export '
    msg2 = msg2.replace('stations, ', '').replace('_t1', '')
    p_export.add_argument('key', help=msg1 + msg2)
    p_export.add_argument('fname', help='filename')

    for p2 in (p_print, p_load, p_plot, p_export):
        msg = 'seed id (only for data or prepdata, e.g. CX.PATCX..BHZ)'
        p2.add_argument('seedid', help=msg, nargs='?', default=SUPPRESS)
        msg = 'day (only for data or prepdata, e.g. 2010-02-03, 2010-035)'
        p2.add_argument('day', help=msg, nargs='?', default=SUPPRESS)
        msg = 'configuration key for correlation (only for prepdata)'
        p2.add_argument('corrid', help=msg, nargs='?', default=SUPPRESS)
    for p2 in (p_correlate, p_stack, p_stretch):
        msg = 'Set chattiness on command line. Up to 3 -v flags are possible'
        p2.add_argument('-v', '--verbose', help=msg, action='count',
                        default=SUPPRESS)
    for p2 in (p_correlate, p_stack, p_stretch):
        msg = ('Number of cores to use (default: all), '
               'only applies to some commands')
        p2.add_argument('-n', '--njobs', default=None, type=int, help=msg)
    msg = ('Run inner loops parallel instead of outer loop '
           '(preproccessing of different stations and correlation of different '
           'pairs versus processing of different days). '
           'Useful for a datset with many stations.')
    p_correlate.add_argument('--parallel-inner-loop', action='store_true',
                             help=msg)
    msg = 'type of plot (a default is chosen for the given key)'
    choices = ('vs_dist', 'wiggle')
    p_plot.add_argument('--plottype', help=msg, choices=choices)
    msg = 'specify some plot options (dictionary in JSON format)'
    p_plot.add_argument('--plot-options', default=SUPPRESS, type=json.loads,
                        help=msg)
    p_plot.add_argument('--show', action='store_true', default=SUPPRESS,
                        help='show figures')
    p_plot.add_argument('--no-show', dest='show',
                        action='store_false', default=SUPPRESS,
                        help='do not show figures')
    msg = 'format supported by ObsPy (default: auto-detected by extension)'
    p_export.add_argument('-f', '--format', help=msg)

    # Get command line arguments and start run function
    args = vars(p.parse_args(args))
    try:
        run(**args)
    except ParseError as ex:
        p.error(ex)
    except ConfigError as ex:
        print(ex, file=sys.stderr)
        sys.exit(1)



# Copyright 2017-2019 Tom Eulenfeld, MIT license
import unittest
import sys

import numpy as np
from obspy import read, read_inventory, UTCDateTime as UTC
from obspy.signal.cross_correlation import correlate, xcorr_max
from scipy.signal import periodogram
from scipy.fftpack import next_fast_len

from yam.correlate import (_fill_array, _downsample_and_shift,
                           _make_same_length,
                           correlate as yam_correlate,
                           correlate_traces, preprocess, spectral_whitening,
                           time_norm)


class TestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        args = sys.argv[1:]
        cls.njobs = int(args[args.index('-n') + 1]) if '-n' in args else None

    def test_fill_array(self):
        data = np.arange(5, dtype=float)
        self.assertIs(_fill_array(data), data)
        data2 = np.copy(data)
        mask = [True, False, False, True, False]
        data2 = _fill_array(data, mask, fill_value=0.)
        self.assertEqual(data2.data[0], 0.)
        self.assertEqual(data2.data[-2], 0.)

    def test_time_norm(self):
        stream = read().select(component='Z')
        stream.filter('highpass', freq=0.5)
        tr1 = stream[0].copy()  # unmasked
        t = stream[0].stats.starttime
        stream.cutout(t + 18, t + 20)
        stream.merge()
        tr2 = stream[0]  # array with masked data
        mask = tr2.data.mask
        tr1_1bit = time_norm(tr1.copy(), '1bit')
        #tr2_1bit = time_norm(np.ma.copy(data2), '1bit', sr)
        tr2_1bit = time_norm(tr2.copy(), '1bit')
        self.assertSetEqual(set(tr1_1bit.data), {-1., 0., 1.})
        self.assertSetEqual(set(tr2_1bit.data._data), {-1., 0., 1.})
        np.testing.assert_equal(tr2_1bit.data.mask, mask)
        np.testing.assert_equal(tr2_1bit.data._data[mask], 0.)

        tr1_clip = time_norm(tr1.copy(), 'clip', clip_value=2)
        tr2_clip = time_norm(tr2.copy(), 'clip', clip_value=2)
        tr1_clip0 = time_norm(tr1.copy(), 'clip', clip_value=2,
                              clip_mode='zero')
        tr2_clip0 = time_norm(tr2.copy(), 'clip', clip_value=2,
                              clip_mode='zero')
        tr2b_clip0 = time_norm(tr2.copy(), 'clip', clip_value=2,
                               clip_mode='mask')
        clip_mask = np.abs(tr1_clip.data) < np.abs(tr1.data)
        with np.errstate(invalid='ignore'):
            clip_mask2 = np.abs(tr2_clip.data) < np.abs(tr2.data)
        np.testing.assert_equal(tr2_clip.data.mask, mask)
        np.testing.assert_equal(tr2_clip0.data.mask, mask)
        self.assertGreater(np.count_nonzero(clip_mask), 0)
        self.assertGreater(np.count_nonzero(clip_mask2), 0)
        np.testing.assert_equal(tr1_clip0.data[clip_mask], 0.)
        np.testing.assert_equal(tr2_clip0.data[clip_mask2], 0.)
        np.testing.assert_array_less(np.abs(tr2_clip.data[clip_mask2]),
                                     np.abs(tr2.data[clip_mask2]))
        self.assertGreater(np.count_nonzero(tr2b_clip0.data.mask),
                           np.count_nonzero(mask))

        tr1_me = time_norm(tr1.copy(), 'mute_envelope', mute_parts=4)
        tr2_me = time_norm(tr2.copy(), 'mute_envelope', mute_parts=4)
        ind = np.abs(tr1.data) > 0.5 * np.max(tr1.data)
        np.testing.assert_equal(tr1_me.data[ind], 0.)
        np.testing.assert_equal(tr2_me.data[ind], 0.)

    def test_spectral_whitening(self):
        stream = read().select(component='Z')
        filter_ = [0.5, 10]
        stream.filter('highpass', freq=0.5)
        tr1 = stream[0].copy()  # unmasked
        t = tr1.stats.starttime
        stream.cutout(t + 18, t + 20)
        stream.merge()
        tr2 = stream[0]  # trace with masked data
        mask = tr2.data.mask

        time_norm(tr1, 'mute_envelope', mute_parts=4)
        time_norm(tr2, 'mute_envelope', mute_parts=4)

        trw1 = spectral_whitening(tr1.copy())
        trw1_f = spectral_whitening(tr1.copy(), filter=filter_)
        trw1_sm = spectral_whitening(tr1.copy(), smooth=5)
        trw1_sm_f = spectral_whitening(tr1.copy(), smooth=5, filter=filter_)
        trw2 = spectral_whitening(tr2.copy())
        trw2_f = spectral_whitening(tr2.copy(), filter=filter_)
        trw2_sm = spectral_whitening(tr2.copy(), smooth=5)
        trw2_sm_f = spectral_whitening(tr2.copy(), smooth=5, filter=filter_)

        np.testing.assert_equal(trw2.data.mask, mask)
        np.testing.assert_equal(trw2_f.data.mask, mask)
        np.testing.assert_equal(trw2_sm.data.mask, mask)
        np.testing.assert_equal(trw2_sm_f.data.mask, mask)

        nfft = next_fast_len(len(tr1.data))
        sr = stream[0].stats.sampling_rate
        f, psd1 = periodogram(tr1.data, sr, nfft=nfft)
        f, psd2 = periodogram(tr2.data, sr, nfft=nfft)
        f, wpsd1 = periodogram(trw1.data, sr, nfft=nfft)
        f, wpsd1_f = periodogram(trw1_f.data, sr, nfft=nfft)
        f, wpsd1_sm = periodogram(trw1_sm.data, sr, nfft=nfft)
        f, wpsd1_sm_f = periodogram(trw1_sm_f.data, sr, nfft=nfft)
        f, wpsd2 = periodogram(trw2.data, sr, nfft=nfft)
        f, wpsd2_f = periodogram(trw2_f.data, sr, nfft=nfft)
        f, wpsd2_sm = periodogram(trw2_sm.data, sr, nfft=nfft)
        f, wpsd2_sm_f = periodogram(trw2_sm_f.data, sr, nfft=nfft)

        ind = f < 10
        psd1 /= np.max(psd1[ind])
        wpsd1 /= np.max(wpsd1[ind])
        wpsd1_f /= np.max(wpsd1_f[ind])
        wpsd1_sm /= np.max(wpsd1_sm[ind])
        wpsd1_sm_f /= np.max(wpsd1_sm_f[ind])
        psd2 /= np.max(psd2[ind])
        wpsd2 /= np.max(wpsd2[ind])
        wpsd2_f /= np.max(wpsd2_f[ind])
        wpsd2_sm /= np.max(wpsd2_sm[ind])
        wpsd2_sm_f /= np.max(wpsd2_sm_f[ind])

        np.testing.assert_allclose(wpsd1[1:-1], 1)
        np.testing.assert_array_less(wpsd1_f[1:] / wpsd1[1:], 1.01)

#        import matplotlib.pyplot as plt
#        ax1 = plt.subplot(221)
#        plt.title('contiguous data')
#        times = tr1.times()
#        plt.plot(times, 20 * tr1.data, label='data')
#        plt.plot(times, trw1.data, label='whitened')
#        plt.plot(times, trw1_f.data, label='+filtered')
#        plt.plot(times, trw1_sm.data, label='+smoothed')
#        plt.plot(times, trw1_sm_f.data, label='+both')
#        plt.xlabel('time (s)')
#        plt.legend()
#
#        plt.subplot(222, sharex=ax1, sharey=ax1)
#        plt.title('gappy data')
#        plt.plot(times, tr2.data / np.max(tr2.data) / 5, label='data')
#        plt.plot(times, trw2.data, label='whitened')
#        plt.plot(times, trw2_f.data, label='+filtered')
#        plt.plot(times, trw2_sm.data, label='+smoothed')
#        plt.plot(times, trw2_sm_f.data, label='+bothed')
#        plt.xlabel('time (s)')
#        plt.legend()
#
#        ax3 = plt.subplot(223)
#        plt.plot(f, psd1, label='data')
#        plt.plot(f, wpsd1, label='whitened')
#        plt.plot(f, wpsd1_f, label='+filtered')
#        plt.plot(f, wpsd1_sm, label='+smoothed')
#        plt.plot(f, wpsd1_sm_f, label='+both')
#        plt.xlabel('frequency (Hz)')
#        plt.legend()
#
#        plt.subplot(224, sharex=ax3, sharey=ax3)
#        plt.plot(f, psd2, label='data')
#        plt.plot(f, wpsd2, label='whitened')
#        plt.plot(f, wpsd2_f, label='+filtered')
#        plt.plot(f, wpsd2_sm, label='+smoothed')
#        plt.plot(f, wpsd2_sm_f, label='+both')
#        plt.xlabel('frequency (Hz)')
#        plt.ylim(0, 1.2)
#        plt.legend()
#        plt.show()

    def test_shift(self):
        tr = read()[0]
        dt = tr.stats.delta
        t = tr.stats.starttime = UTC('2018-01-01T00:00:10.000000Z')
        tr2 = tr.copy()
        _downsample_and_shift(tr2)
        self.assertEqual(tr2, tr)

        tr2 = tr.copy()
        tr2.stats.starttime = t + 0.1 * dt
        _downsample_and_shift(tr2)
        self.assertEqual(tr2.stats.starttime, t)

        tr2 = tr.copy()
        tr2.stats.starttime = t - 0.1 * dt
        _downsample_and_shift(tr2)
        self.assertEqual(tr2.stats.starttime, t)

        tr2 = tr.copy()
        tr2.stats.starttime = t - 0.49 * dt
        _downsample_and_shift(tr2)
        self.assertEqual(tr2.stats.starttime, t)

        tr2 = tr.copy()
        tr2.stats.starttime = t - 0.0001 * dt
        _downsample_and_shift(tr2)
        self.assertEqual(tr2.stats.starttime, t)

        # shift cumulatively by +1 sample
        tr2 = tr.copy()
        tr2.stats.starttime += 0.3 * dt
        _downsample_and_shift(tr2)
        tr2.stats.starttime += 0.3 * dt
        _downsample_and_shift(tr2)
        tr2.stats.starttime += 0.4 * dt
        _downsample_and_shift(tr2)
        self.assertEqual(tr2.stats.starttime, t)
        np.testing.assert_allclose(tr2.data[201:-200], tr.data[200:-201],
                                   rtol=1e-2, atol=1)
        cc = correlate(tr2.data, tr.data, 1000)
        shift, cc_max = xcorr_max(cc)
        self.assertEqual(shift, 1)
        self.assertGreater(cc_max, 0.995)

        # shift cumulatively by -1 sample
        tr2 = tr.copy()
        tr2.stats.starttime -= 0.3 * dt
        _downsample_and_shift(tr2)
        tr2.stats.starttime -= 0.3 * dt
        _downsample_and_shift(tr2)
        tr2.stats.starttime -= 0.4 * dt
        _downsample_and_shift(tr2)
        self.assertEqual(tr2.stats.starttime, t)
        np.testing.assert_allclose(tr2.data[200:-201], tr.data[201:-200],
                                   rtol=1e-2, atol=2)
        cc = correlate(tr2.data, tr.data, 1000)
        shift, cc_max = xcorr_max(cc)
        self.assertEqual(shift, -1)
        self.assertGreater(cc_max, 0.995)

    def test_downsample_and_shift(self):
        tr = read()[0]
        t = tr.stats.starttime = UTC('2018-01-01T00:00:10.000000Z')
        # decimate
        tr2 = _downsample_and_shift(tr.copy(), 50.)
        self.assertEqual(tr2.stats.sampling_rate, 50)
        # interpolate
        tr2 = _downsample_and_shift(tr.copy(), 40.)
        self.assertEqual(tr2.stats.sampling_rate, 40)
        # decimate and time shift
        tr2 = tr.copy()
        tr2.stats.starttime += 0.002
        tr2 = _downsample_and_shift(tr2, 50.)
        self.assertEqual(tr2.stats.sampling_rate, 50)
        self.assertEqual(tr2.stats.starttime, t)
        tr2 = tr.copy()
        tr2.stats.starttime -= 0.002
        tr2 = _downsample_and_shift(tr2, 50.)
        self.assertEqual(tr2.stats.sampling_rate, 50)
        self.assertEqual(tr2.stats.starttime, t)
        # interpolate and time shift
        tr2 = tr.copy()
        tr2.stats.starttime += 0.002
        tr2 = _downsample_and_shift(tr2, 40.)
        self.assertEqual(tr2.stats.sampling_rate, 40)
        self.assertEqual(tr2.stats.starttime - tr2.stats.delta, t)
        tr2 = tr.copy()
        tr2.stats.starttime -= 0.002
        tr2 = _downsample_and_shift(tr2, 40.)
        self.assertEqual(tr2.stats.sampling_rate, 40)
        self.assertEqual(tr2.stats.starttime, t)

    def test_preprocess(self):
        stream = read()
        day = UTC('2018-01-02')
        for tr in stream:
            tr.stats.starttime = day
        tr = stream[1]
        tr.id = 'GR.FUR..BH' + tr.stats.channel[-1]
        tr.stats.sampling_rate = 80.
        tr = stream[2]
        tr.id = 'GR.WET..BH' + tr.stats.channel[-1]
        tr.stats.sampling_rate = 50.
        stream = stream.cutout(day + 0.01, day + 10)
        stream = stream.cutout(day + 14, day + 16.05)
        norm = ('clip', 'spectral_whitening', 'mute_envelope', '1bit')
        # see https://docs.scipy.org/doc/numpy-1.13.0/release.html#
        # assigning-to-slices-views-of-maskedarray
        ignore_msg = r'setting an item on a masked array which has a shared'
        with np.warnings.catch_warnings():
            np.warnings.filterwarnings('ignore', ignore_msg)
            preprocess(stream, day=day, inventory=read_inventory(),
                       remove_response=True,
                       filter=None,
                       normalization=norm,
                       time_norm_options=None,
                       spectral_whitening_options=None,
                       decimate=5)
        for tr in stream:
            self.assertEqual(tr.stats.sampling_rate, 10)
        for tr in stream:
            self.assertEqual(set(tr.data._data), {-1, 0, 1})
            mask = np.ma.getmask(tr.data)
            np.testing.assert_equal(tr.data[mask]._data, 0)
            self.assertGreater(np.count_nonzero(mask), 0)
        self.assertEqual(len(stream), 3)

    def test_correlate_traces(self):
        stream = read().sort()
        tr = correlate_traces(stream[0], stream[1])
        shift, corr = xcorr_max(tr.data)
        self.assertLess(abs(shift / len(tr)), 0.01)
        self.assertGreater(abs(corr), 0.2)
        self.assertEqual(tr.id, 'RJOB.EHE.RJOB.EHN')

    def test_correlate(self):
        stream = read()
        stream2 = stream.copy()
        stream3 = stream.copy()
        for tr in stream2:
            tr.id = 'GR.FUR..BH' + tr.stats.channel[-1]
            tr.stats.sampling_rate = 80.
        for tr in stream3:
            tr.id = 'GR.WET..BH' + tr.stats.channel[-1]
            tr.stats.sampling_rate = 50.
        stream = stream + stream2 + stream3
        day = UTC('2018-01-02')
        for tr in stream:
            tr.stats.starttime = day
        # create some gaps
        stream = stream.cutout(day + 0.01, day + 10)
        stream = stream.cutout(day + 14, day + 16.05)

        # prepare mock objects for call to yam_correlate
        def data(starttime, endtime, **kwargs):
            return stream.select(**kwargs).slice(starttime, endtime)
        io = {'data': data, 'data_format': None, 'inventory': read_inventory()}
        res = yam_correlate(io, day, 'outkey', keep_correlations=True)
        self.assertEqual(len(res['corr']), 6)
        # by default only 'ZZ' combinations
        for tr in res['corr']:
            self.assertEqual(tr.stats.station[-1], 'Z')
            self.assertEqual(tr.stats.channel[-1], 'Z')
            if len(set(tr.id.split('.'))) == 2:  # autocorr
                np.testing.assert_allclose(xcorr_max(tr.data), (0, 1.))

        res = yam_correlate(
            io, day, 'outkey',
            station_combinations=('GR.FUR-GR.WET', 'RJOB-RJOB'),
            component_combinations=('ZZ', 'NE', 'NR'),
            keep_correlations=True,
            stack='1d', njobs=self.njobs)
        self.assertEqual(len(res['corr']), 7)
        self.assertEqual(len(res['stack']), 7)
        ids = ['RJOB.EHE.RJOB.EHN', 'RJOB.EHZ.RJOB.EHZ',
               'FUR.BHE.WET.BHN', 'FUR.BHN.WET.BHE',
               'FUR.BHR.WET.BHN', 'FUR.BHN.WET.BHR',
               'FUR.BHZ.WET.BHZ']
        for tr in res['corr']:
            self.assertIn(tr.id, ids)
            if len(set(tr.id.split('.'))) == 2:  # autocorr
                np.testing.assert_allclose(xcorr_max(tr.data), (0, 1.))

        res = yam_correlate(
            io, day, 'outkey', only_auto_correlation=True,
            station_combinations=('GR.FUR-GR.WET', 'RJOB-RJOB'),
            component_combinations=['ZN', 'RT'], njobs=self.njobs,
            keep_correlations=True,
            remove_response=True)
        self.assertEqual(len(res['corr']), 1)
        tr = res['corr'][0]
        self.assertEqual(tr.stats.station[-1], 'N')
        self.assertEqual(tr.stats.channel[-1], 'Z')

        stream.traces = [tr for tr in stream if tr.stats.channel[-1] != 'N']
        res = yam_correlate(
            io, day, 'outkey',
            station_combinations=('GR.FUR-GR.WET', 'RJOB-RJOB'),
            component_combinations=('NT', 'NR'), discard=0.0,
            keep_correlations=True)
        self.assertEqual(res, None)

    def test_make_same_length(self):
        tr1 = read()[0]
        tr1.stats.sampling_rate = 1
        utc = tr1.stats.starttime
        tr2 = tr1.copy()
        tr2.stats.starttime -= 0.1
        for dt in np.linspace(0, 1, 101):
            tr1a = tr1.slice(utc + dt, utc + 3.7 + dt)
            tr2a = tr2.slice(utc + dt, utc + 3.7 + dt)
            _make_same_length(tr1a, tr2a)
            self.assertEqual(len(tr1a), len(tr2a))
            self.assertLess(tr1a.stats.starttime - tr2a.stats.starttime, 0.5)


def suite():
    return unittest.makeSuite(TestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# Copyright 2018-2019 Tom Eulenfeld, MIT license

import os.path
import tempfile
import unittest

import numpy as np
from obspy import Stream, Trace, UTCDateTime as UTC

from yam.stretch import stretch
from yam.commands import write_dict, read_dicts


class TestCase(unittest.TestCase):

    def test_stretch(self):
        h = {'sampling_rate': 100}
        h['network1'] = h['network2'] = 'NET'
        h['station1'] = h['station2'] = h['network'] = h['location'] = 'STA'
        h['location1'] = h['location2'] = ''
        h['channel1'] = h['channel2'] = h['location'] = h['channel'] = 'HHZ'
        h['dist'] = h['azi'] = h['baz'] = 0
        vel_changes = [0, 1, -1]
        traces = []
        dt = 24 * 3600
        t0 = UTC()
        for i, v in enumerate(vel_changes):
            mul = 1 + v / 100
            t = np.linspace(-10 * mul, 10 * mul, 10001)
            data = np.cos(2 * np.pi * t)
            h['starttime'] = t0 + i * dt
            tr = Trace(data, header=h)
            traces.append(tr)
        d = stretch(Stream(traces), max_stretch=1.1, num_stretch=2201,
                    tw=(1, 5), sides='both', reftr=traces[0])
        expect = np.array(vel_changes)
        np.testing.assert_allclose(d['velchange_vs_time'], expect)
        np.testing.assert_allclose(d['corr_vs_time'], (1, 1, 1))
        self.assertAlmostEqual(d['velchange_values'][-1], 1.1)
        self.assertEqual(len(d['velchange_values']), 2201)
        # test writing and reading
        with tempfile.TemporaryDirectory(prefix='yam_') as tmpdir:
            fname = os.path.join(tmpdir, 'stretch.h5')
            d['attrs']['key'] = 'test'
            write_dict(d, fname)
            d2 = read_dicts(fname)[0]
            for key in d:
                if key == 'sim_mat':
                    np.testing.assert_allclose(d2[key], d[key], rtol=1e-3)
                elif isinstance(d2[key], np.ndarray):
                    np.testing.assert_equal(d2[key], d[key])
                else:
                    self.assertEqual(d2[key], d[key])
            d2['attrs']['key'] = 'test2'
            write_dict(d2, fname)


def suite():
    return unittest.makeSuite(TestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# Copyright 2017-2019, Tom Eulenfeld, MIT license
"""
Stretch correlations

The results are returned in a dictionary with the following entries:

:times: strings of starttimes of the traces (1D array, length ``N1``)
:velchange_values: velocity changes (%) corresponding to the used stretching
    factors (assuming a homogeneous velocity change, 1D array, length ``N2``)
:tw: used lag time window
:sim_mat: similarity matrices (2D array, dimension ``(N1, N2)``)
:velchange_vs_time: velocity changes (%) as a function of time
    (value of highest correlation/similarity for each time, length ``N1``)
:corr_vs_time: correlation values as a function of time
    (value of highest correlation/similarity for each time, length ``N1``)
:attrs: dictionary with metadata
    (e.g. network, station, channel information of both stations,
    inter-station distance and parameters passed to the stretching function)

|
"""

import logging
import functools
import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline
from warnings import warn

from yam.util import _corr_id, _trim, _trim_time_period
import yam.stack

log = logging.getLogger('yam.stretch')


def _intersect_sorted(l1, l2):
    i = 0
    res = []
    for el in l1:
        while l2[i] < el:
            i += 1
        if l2[i] == el:
            res.append(el)
            i += 1
    return res


def _index_sorted(l1, l2):
    index = np.empty(len(l1), dtype=bool)
    i = 0
    for j, el in enumerate(l1):
        if len(l2) <= i:
            index[j:] = False
            break
        if l2[i] == el:
            index[j] = True
            i += 1
        else:
            index[j] = False
    return index


def _update_result(res):
    sim_mat = res['sim_mat']
    res['corr_vs_time'] = np.max(sim_mat, axis=1)
    argmax = np.argmax(sim_mat, axis=1)
    res['velchange_vs_time'] = res['velchange_values'][argmax]


def join_dicts(dicts):
    """Join list of dictionaries with stretching results"""
    # TODO: write test for this function
    if len(dicts) == 0:
        return
    elif len(dicts) == 1:
        return dicts[0]
    dicts = sorted(dicts, key=lambda d: d['times'][0])
    dim1 = sum(len(d['times']) for d in dicts)
    d = dicts[0]
    dim2 = len(d['velchange_values'])
    dtype = d['sim_mat'].dtype
    res = {'sim_mat': np.empty((dim1, dim2), dtype=dtype),
           'velchange_values': d['velchange_values'],
           'times': np.empty(dim1, dtype=d['times'].dtype),
           'velchange_vs_time': np.empty(dim1, dtype=dtype),
           'corr_vs_time': np.empty(dim1, dtype=dtype),
           'lag_time_windows': d['lag_time_windows'],
           'attrs': d['attrs']}
    res['attrs']['endtime'] = dicts[-1]['attrs']['endtime']
    i = 0
    for d in dicts:
        j = i + len(d['times'])
        res['sim_mat'][i:j, :] = d['sim_mat']
        res['times'][i:j] = d['times']
        res['velchange_vs_time'][i:j] = d['velchange_vs_time']
        res['corr_vs_time'][i:j] = d['corr_vs_time']
        i = j
    return res


def average_dicts(dicts):
    """Average list of dictionaries with stretching results"""
    # TODO: write test for this function
    if len(dicts) == 0:
        return
    elif len(dicts) == 1:
        return dicts[0]
    times = functools.reduce(_intersect_sorted, [d['times'] for d in dicts])
    d = dicts[0]
    sim_mat = np.mean(
        [d['sim_mat'][_index_sorted(d['times'], times), :] for d in dicts],
        axis=0)
    res = {'sim_mat': sim_mat,
           'velchange_values': d['velchange_values'],
           'times': np.array(times),
           'tw': d['tw'],
           'attrs': d['attrs']}
    res['attrs']['channel1'] = '???'
    res['attrs']['channel2'] = '???'
    _update_result(res)
    return res


def _stretch_helper(data, refdata, mask, stretch_factor):
    N1, N2 = data.shape
    N3 = len(stretch_factor)
    assert N2 == len(refdata)
    # create a spline object for the reference trace
    # and evaluate the spline object at stretched times
    time_index_max = (N2 - 1) / 2
    time_idx = np.linspace(-time_index_max, time_index_max, N2)
    ref_tr_spline = InterpolatedUnivariateSpline(time_idx, refdata)
    ref_stretch = np.empty((N3, N2))
    for i in range(N3):
        ref_stretch[i, :] = ref_tr_spline(time_idx * stretch_factor[i])
    # correlate, normalize and return
    cc = (data * mask) @ np.transpose(ref_stretch * mask)
    sq1 = np.sum((data * mask) ** 2, axis=1)
    sq2 = np.sum((ref_stretch * mask) ** 2, axis=1)
    norm = (sq1[:, np.newaxis] * sq2) ** 0.5
    sim_mat = cc / norm
    sim_mat[np.isnan(sim_mat)] = 0
    assert sim_mat.shape == (N1, N3)
    return sim_mat


def stretch(stream, max_stretch, num_stretch, tw, tw_relative=None,
            reftr=None, sides='both', max_lag=None, time_period=None
            ):
    """
    Stretch traces in stream and return dictionary with results

    See e.g. Richter et al. (2015) for a description of the procedure.

    :param stream: |Stream| object with correlations
    :param float max_stretch: stretching range in percent
    :param int num_stretch: number of values in stretching vector
    :param tw: definition of the time window in the correlation --
        tuple of length 2 with start and end time in seconds (positive)
    :param tw_relative: time windows can be defined relative to a
        velocity, default None or 0 -- time windows relative to zero lag time,
        otherwise velocity is given in km/s
    :param reftr: reference trace, by default the stack of stream is used
        as reference
    :param sides: one of left, right, both
    :param max_lag: max lag time in seconds, stream is trimmed to
        ``(-max_lag, max_lag)`` before stretching
    :param time_period: use correlations only from this time span
        (tuple of dates)
    """
    if len(stream) <= 1:
        log.warning('For stretch the stream must have a minimum length of 2')
        return
    ids = {_corr_id(tr) for tr in stream}
    if None in ids:
        ids.discard(None)
        stream.traces = [tr for tr in stream if _corr_id(tr) is not None]
    if len(ids) != 1:
        warn('Different ids in stream: %s' % ids)
    _trim_time_period(stream, time_period)
    stream.sort()
    tr0 = stream[0]
    if max_lag is not None:
        for tr in stream:
            _trim(tr, (-max_lag, max_lag))
    rel = 0 if tw_relative is None else tr0.stats.dist / 1000 / tw_relative
    twabs = rel + np.array(tw)
    starttime = tr0.stats.starttime
    mid = starttime + (tr0.stats.endtime - starttime) / 2
    times = tr0.times(reftime=mid)
    data = np.array([tr.data for tr in stream])
    data[np.isnan(data)] = 0  # bug fix
    data[np.isinf(data)] = 0
    if reftr is None:
        reftr = yam.stack.stack(stream)[0]
    stretch_vector_perc = np.linspace(-max_stretch, max_stretch, num_stretch)
    stretch_factor = 1 + stretch_vector_perc / 100
    # MIIC approximation:
    #stretch_factor = np.exp(stretch_vector_percent / 100)

    mask1 = np.logical_and(times >= twabs[0], times <= twabs[1])
    mask2 = np.logical_and(times <= -twabs[0], times >= -twabs[1])
    if sides == 'left':
        mask = mask1
    elif sides == 'right':
        mask = mask2
    elif sides == 'both':
        mask = np.logical_or(mask1, mask2)
    else:
        raise ValueError('sides = %s not a valid option' % sides)
    sim_mat = _stretch_helper(data, reftr.data, mask, stretch_factor)
    times = np.array([str(tr.stats.starttime)[:19] for tr in stream],
                     dtype='S19')
    result = {'sim_mat': sim_mat,
              'velchange_values': stretch_vector_perc,
              'times': times,
              'tw': twabs,
              'attrs': {'num_stretch': num_stretch,
                        'max_stretch': max_stretch,
                        'sides': sides,
                        'starttime': stream[0].stats.starttime,
                        'endtime': stream[-1].stats.starttime
                        }
              }
    _update_result(result)
    for k in ('network1', 'network2', 'station1', 'station2',
              'location1', 'location2', 'channel1', 'channel2',
              'sampling_rate', 'dist', 'azi', 'baz'):
        if k in tr0.stats:
            result['attrs'][k] = tr0.stats[k]
    return result
